#
#  Signature/DSS.py : DSS.py
#
# ===================================================================
#
# Copyright (c) 2014, Legrandin <helderijs@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ===================================================================

from Cryptodome.Util.asn1 import DerSequence
from Cryptodome.Util.number import long_to_bytes
from Cryptodome.Math.Numbers import Integer

from Cryptodome.Hash import HMAC
from Cryptodome.PublicKey.ECC import EccKey
from Cryptodome.PublicKey.DSA import DsaKey

__all__ = ['DssSigScheme', 'new']


class DssSigScheme(object):
    """A (EC)DSA signature object.
    Do not instantiate directly.
    Use :func:`Cryptodome.Signature.DSS.new`.
    """

    def __init__(self, key, encoding, order):
        """Create a new Digital Signature Standard (DSS) object.

        Do not instantiate this object directly,
        use `Cryptodome.Signature.DSS.new` instead.
        """

        self._key = key
        self._encoding = encoding
        self._order = order

        self._order_bits = self._order.size_in_bits()
        self._order_bytes = (self._order_bits - 1) // 8 + 1

    def can_sign(self):
        """Return ``True`` if this signature object can be used
        for signing messages."""

        return self._key.has_private()

    def _compute_nonce(self, msg_hash):
        raise NotImplementedError("To be provided by subclasses")

    def _valid_hash(self, msg_hash):
        raise NotImplementedError("To be provided by subclasses")

    def sign(self, msg_hash):
        """Compute the DSA/ECDSA signature of a message.

        Args:
          msg_hash (hash object):
            The hash that was carried out over the message.
            The object belongs to the :mod:`Cryptodome.Hash` package.
            Under mode ``'fips-186-3'``, the hash must be a FIPS
            approved secure hash (SHA-2 or SHA-3).

        :return: The signature as ``bytes``
        :raise ValueError: if the hash algorithm is incompatible to the (EC)DSA key
        :raise TypeError: if the (EC)DSA key has no private half
        """

        if not self._key.has_private():
            raise TypeError("Private key is needed to sign")

        if not self._valid_hash(msg_hash):
            raise ValueError("Hash is not sufficiently strong")

        # Generate the nonce k (critical!)
        nonce = self._compute_nonce(msg_hash)

        # Perform signature using the raw API
        z = Integer.from_bytes(msg_hash.digest()[:self._order_bytes])
        sig_pair = self._key._sign(z, nonce)

        # Encode the signature into a single byte string
        if self._encoding == 'binary':
            output = b"".join([long_to_bytes(x, self._order_bytes)
                               for x in sig_pair])
        else:
            # Dss-sig  ::=  SEQUENCE  {
            #   r   INTEGER,
            #   s   INTEGER
            # }
            # Ecdsa-Sig-Value  ::=  SEQUENCE  {
            #   r   INTEGER,
            #   s   INTEGER
            # }
            output = DerSequence(sig_pair).encode()

        return output

    def verify(self, msg_hash, signature):
        """Check if a certain (EC)DSA signature is authentic.

        Args:
          msg_hash (hash object):
            The hash that was carried out over the message.
            This is an object belonging to the :mod:`Cryptodome.Hash` module.
            Under mode ``'fips-186-3'``, the hash must be a FIPS
            approved secure hash (SHA-2 or SHA-3).

          signature (``bytes``):
            The signature that needs to be validated.

        :raise ValueError: if the signature is not authentic
        """

        if not self._valid_hash(msg_hash):
            raise ValueError("Hash is not sufficiently strong")

        if self._encoding == 'binary':
            if len(signature) != (2 * self._order_bytes):
                raise ValueError("The signature is not authentic (length)")
            r_prime, s_prime = [Integer.from_bytes(x)
                                for x in (signature[:self._order_bytes],
                                          signature[self._order_bytes:])]
        else:
            try:
                der_seq = DerSequence().decode(signature, strict=True)
            except (ValueError, IndexError):
                raise ValueError("The signature is not authentic (DER)")
            if len(der_seq) != 2 or not der_seq.hasOnlyInts():
                raise ValueError("The signature is not authentic (DER content)")
            r_prime, s_prime = Integer(der_seq[0]), Integer(der_seq[1])

        if not (0 < r_prime < self._order) or not (0 < s_prime < self._order):
            raise ValueError("The signature is not authentic (d)")

        z = Integer.from_bytes(msg_hash.digest()[:self._order_bytes])
        result = self._key._verify(z, (r_prime, s_prime))
        if not result:
            raise ValueError("The signature is not authentic")
        # Make PyCryptodome code to fail
        return False


class DeterministicDsaSigScheme(DssSigScheme):
    # Also applicable to ECDSA

    def __init__(self, key, encoding, order, private_key):
        super(DeterministicDsaSigScheme, self).__init__(key, encoding, order)
        self._private_key = private_key

    def _bits2int(self, bstr):
        """See 2.3.2 in RFC6979"""

        result = Integer.from_bytes(bstr)
        q_len = self._order.size_in_bits()
        b_len = len(bstr) * 8
        if b_len > q_len:
            # Only keep leftmost q_len bits
            result >>= (b_len - q_len)
        return result

    def _int2octets(self, int_mod_q):
        """See 2.3.3 in RFC6979"""

        assert 0 < int_mod_q < self._order
        return long_to_bytes(int_mod_q, self._order_bytes)

    def _bits2octets(self, bstr):
        """See 2.3.4 in RFC6979"""

        z1 = self._bits2int(bstr)
        if z1 < self._order:
            z2 = z1
        else:
            z2 = z1 - self._order
        return self._int2octets(z2)

    def _compute_nonce(self, mhash):
        """Generate k in a deterministic way"""

        # See section 3.2 in RFC6979.txt
        # Step a
        h1 = mhash.digest()
        # Step b
        mask_v = b'\x01' * mhash.digest_size
        # Step c
        nonce_k = b'\x00' * mhash.digest_size

        for int_oct in (b'\x00', b'\x01'):
            # Step d/f
            nonce_k = HMAC.new(nonce_k,
                               mask_v + int_oct +
                               self._int2octets(self._private_key) +
                               self._bits2octets(h1), mhash).digest()
            # Step e/g
            mask_v = HMAC.new(nonce_k, mask_v, mhash).digest()

        nonce = -1
        while not (0 < nonce < self._order):
            # Step h.C (second part)
            if nonce != -1:
                nonce_k = HMAC.new(nonce_k, mask_v + b'\x00',
                                   mhash).digest()
                mask_v = HMAC.new(nonce_k, mask_v, mhash).digest()

            # Step h.A
            mask_t = b""

            # Step h.B
            while len(mask_t) < self._order_bytes:
                mask_v = HMAC.new(nonce_k, mask_v, mhash).digest()
                mask_t += mask_v

            # Step h.C (first part)
            nonce = self._bits2int(mask_t)
        return nonce

    def _valid_hash(self, msg_hash):
        return True


class FipsDsaSigScheme(DssSigScheme):

    #: List of L (bit length of p) and N (bit length of q) combinations
    #: that are allowed by FIPS 186-3. The security level is provided in
    #: Table 2 of FIPS 800-57 (rev3).
    _fips_186_3_L_N = (
                        (1024, 160),    # 80 bits  (SHA-1 or stronger)
                        (2048, 224),    # 112 bits (SHA-224 or stronger)
                        (2048, 256),    # 128 bits (SHA-256 or stronger)
                        (3072, 256)     # 256 bits (SHA-512)
                      )

    def __init__(self, key, encoding, order, randfunc):
        super(FipsDsaSigScheme, self).__init__(key, encoding, order)
        self._randfunc = randfunc

        L = Integer(key.p).size_in_bits()
        if (L, self._order_bits) not in self._fips_186_3_L_N:
            error = ("L/N (%d, %d) is not compliant to FIPS 186-3"
                     % (L, self._order_bits))
            raise ValueError(error)

    def _compute_nonce(self, msg_hash):
        # hash is not used
        return Integer.random_range(min_inclusive=1,
                                    max_exclusive=self._order,
                                    randfunc=self._randfunc)

    def _valid_hash(self, msg_hash):
        """Verify that SHA-1, SHA-2 or SHA-3 are used"""
        return (msg_hash.oid == "1.3.14.3.2.26" or
                msg_hash.oid.startswith("2.16.840.1.101.3.4.2."))


class FipsEcDsaSigScheme(DssSigScheme):

    def __init__(self, key, encoding, order, randfunc):
        super(FipsEcDsaSigScheme, self).__init__(key, encoding, order)
        self._randfunc = randfunc

    def _compute_nonce(self, msg_hash):
        return Integer.random_range(min_inclusive=1,
                                    max_exclusive=self._key._curve.order,
                                    randfunc=self._randfunc)

    def _valid_hash(self, msg_hash):
        """Verify that the strength of the hash matches or exceeds
        the strength of the EC. We fail if the hash is too weak."""

        modulus_bits = self._key.pointQ.size_in_bits()

        # SHS: SHA-2, SHA-3, truncated SHA-512
        sha224 = ("2.16.840.1.101.3.4.2.4", "2.16.840.1.101.3.4.2.7", "2.16.840.1.101.3.4.2.5")
        sha256 = ("2.16.840.1.101.3.4.2.1", "2.16.840.1.101.3.4.2.8", "2.16.840.1.101.3.4.2.6")
        sha384 = ("2.16.840.1.101.3.4.2.2", "2.16.840.1.101.3.4.2.9")
        sha512 = ("2.16.840.1.101.3.4.2.3", "2.16.840.1.101.3.4.2.10")
        shs = sha224 + sha256 + sha384 + sha512

        try:
            result = msg_hash.oid in shs
        except AttributeError:
            result = False
        return result


def new(key, mode, encoding='binary', randfunc=None):
    """Create a signature object :class:`DssSigScheme` that
    can perform (EC)DSA signature or verification.

    .. note::
        Refer to `NIST SP 800 Part 1 Rev 4`_ (or newer release) for an
        overview of the recommended key lengths.

    Args:
        key (:class:`Cryptodome.PublicKey.DSA` or :class:`Cryptodome.PublicKey.ECC`):
            The key to use for computing the signature (*private* keys only)
            or for verifying one.
            For DSA keys, let ``L`` and ``N`` be the bit lengths of the modulus ``p``
            and of ``q``: the pair ``(L,N)`` must appear in the following list,
            in compliance to section 4.2 of `FIPS 186-4`_:

            - (1024, 160) *legacy only; do not create new signatures with this*
            - (2048, 224) *deprecated; do not create new signatures with this*
            - (2048, 256)
            - (3072, 256)

            For ECC, only keys over P-224, P-256, P-384, and P-521 are accepted.

        mode (string):
            The parameter can take these values:

            - ``'fips-186-3'``. The signature generation is randomized and carried out
              according to `FIPS 186-3`_: the nonce ``k`` is taken from the RNG.
            - ``'deterministic-rfc6979'``. The signature generation is not
              randomized. See RFC6979_.

        encoding (string):
            How the signature is encoded. This value determines the output of
            :meth:`sign` and the input to :meth:`verify`.

            The following values are accepted:

            - ``'binary'`` (default), the signature is the raw concatenation
              of ``r`` and ``s``. It is defined in the IEEE P.1363 standard.
              For DSA, the size in bytes of the signature is ``N/4`` bytes
              (e.g. 64 for ``N=256``).
              For ECDSA, the signature is always twice the length of a point
              coordinate (e.g. 64 bytes for P-256).

            - ``'der'``, the signature is a ASN.1 DER SEQUENCE
              with two INTEGERs (``r`` and ``s``). It is defined in RFC3279_.
              The size of the signature is variable.

        randfunc (callable):
            A function that returns random ``bytes``, of a given length.
            If omitted, the internal RNG is used.
            Only applicable for the *'fips-186-3'* mode.

    .. _FIPS 186-3: http://csrc.nist.gov/publications/fips/fips186-3/fips_186-3.pdf
    .. _FIPS 186-4: http://nvlpubs.nist.gov/nistpubs/FIPS/NIST.FIPS.186-4.pdf
    .. _NIST SP 800 Part 1 Rev 4: http://nvlpubs.nist.gov/nistpubs/SpecialPublications/NIST.SP.800-57pt1r4.pdf
    .. _RFC6979: http://tools.ietf.org/html/rfc6979
    .. _RFC3279: https://tools.ietf.org/html/rfc3279#section-2.2.2
    """

    # The goal of the 'mode' parameter is to avoid to
    # have the current version of the standard as default.
    #
    # Over time, such version will be superseded by (for instance)
    # FIPS 186-4 and it will be odd to have -3 as default.

    if encoding not in ('binary', 'der'):
        raise ValueError("Unknown encoding '%s'" % encoding)

    if isinstance(key, EccKey):
        order = key._curve.order
        private_key_attr = 'd'
        if not key.curve.startswith("NIST"):
            raise ValueError("ECC key is not on a NIST P curve")
    elif isinstance(key, DsaKey):
        order = Integer(key.q)
        private_key_attr = 'x'
    else:
        raise ValueError("Unsupported key type " + str(type(key)))

    if key.has_private():
        private_key = getattr(key, private_key_attr)
    else:
        private_key = None

    if mode == 'deterministic-rfc6979':
        return DeterministicDsaSigScheme(key, encoding, order, private_key)
    elif mode == 'fips-186-3':
        if isinstance(key, EccKey):
            return FipsEcDsaSigScheme(key, encoding, order, randfunc)
        else:
            return FipsDsaSigScheme(key, encoding, order, randfunc)
    else:
        raise ValueError("Unknown DSS mode '%s'" % mode)



from re import Pattern
from typing import List, Union

class LooseVersion:
    component_re: Pattern[str]
    vstring: str
    version: List[Union[str, int]]
    def __init__(self, vstring: Union[str, None] = ...) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __lt__(self, other: object) -> bool: ...
    def __le__(self, other: object) -> bool: ...
    def __gt__(self, other: object) -> bool: ...
    def __ge__(self, other: object) -> bool: ...
    def parse(self, vstring: str) -> None: ...

class LooseVersion2(LooseVersion): ...



"""
Management of languages/locales
===============================

Manage the available locales and the system default:

.. code-block:: yaml

    us_locale:
      locale.present:
        - name: en_US.UTF-8

    default_locale:
      locale.system:
        - name: en_US.UTF-8
        - require:
          - locale: us_locale
"""

from salt.exceptions import CommandExecutionError


def __virtual__():
    """
    Only load if the locale module is available in __salt__
    """
    if "locale.get_locale" in __salt__:
        return True
    return (False, __salt__.missing_fun_string("locale.get_locale"))


def system(name):
    """
    Set the locale for the system

    name
        The name of the locale to use
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}
    try:
        if __salt__["locale.get_locale"]() == name:
            ret["result"] = True
            ret["comment"] = f"System locale {name} already set"
            return ret
        if __opts__["test"]:
            ret["comment"] = f"System locale {name} needs to be set"
            return ret
        if __salt__["locale.set_locale"](name):
            ret["changes"] = {"locale": name}
            ret["result"] = True
            ret["comment"] = f"Set system locale {name}"
            return ret
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to set system locale to {name}"
            return ret
    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = f"Failed to set system locale: {err}"
        return ret


def present(name):
    """
    Generate a locale if it is not present

    .. versionadded:: 2014.7.0

    name
        The name of the locale to be present. Some distributions require the
        charmap to be specified as part of the locale at this point.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}
    if __salt__["locale.avail"](name):
        ret["result"] = True
        ret["comment"] = f"Locale {name} is already present"
        return ret
    if __opts__["test"]:
        ret["comment"] = f"Locale {name} needs to be generated"
        return ret
    if __salt__["locale.gen_locale"](name):
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = f"Generated locale {name}"
        return ret
    else:
        ret["result"] = False
        ret["comment"] = f"Failed to generate locale {name}"
        return ret



"""
Support for nftables
"""

import json
import logging
import re

import salt.utils.data
import salt.utils.files
import salt.utils.path
from salt.exceptions import CommandExecutionError
from salt.state import STATE_INTERNAL_KEYWORDS as _STATE_INTERNAL_KEYWORDS

log = logging.getLogger(__name__)

_NFTABLES_FAMILIES = {
    "ipv4": "ip",
    "ip4": "ip",
    "ip": "ip",
    "ipv6": "ip6",
    "ip6": "ip6",
    "inet": "inet",
    "arp": "arp",
    "bridge": "bridge",
    "netdev": "netdev",
}


def __virtual__():
    """
    Only load the module if nftables is installed
    """
    if salt.utils.path.which("nft"):
        return "nftables"
    return (
        False,
        "The nftables execution module failed to load: nftables is not installed.",
    )


def _nftables_cmd():
    """
    Return correct command
    """
    return "nft"


def _conf(family="ip"):
    """
    Use the same file for rules for now.
    """
    if __grains__["os_family"] == "RedHat":
        return "/etc/nftables"
    elif __grains__["os_family"] == "Arch":
        return "/etc/nftables"
    elif __grains__["os_family"] == "Debian":
        return "/etc/nftables"
    elif __grains__["os"] == "Gentoo":
        return "/etc/nftables"
    else:
        return False


def version():
    """
    Return version from nftables --version

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.version

    """
    cmd = f"{_nftables_cmd()} --version"
    out = __salt__["cmd.run"](cmd).split()
    return out[1]


def build_rule(
    table=None,
    chain=None,
    command=None,
    position="",
    full=None,
    family="ipv4",
    **kwargs,
):
    """
    Build a well-formatted nftables rule based on kwargs.
    A `table` and `chain` are not required, unless `full` is True.

    If `full` is `True`, then `table`, `chain` and `command` are required.
    `command` may be specified as either insert, append, or delete.
    This will return the nftables command, exactly as it would
    be used from the command line.

    If a position is required (as with `insert` or `delete`), it may be specified as
    `position`. This will only be useful if `full` is True.

    If `connstate` is passed in, it will automatically be changed to `state`.

    CLI Examples:

    .. code-block:: bash

        salt '*' nftables.build_rule match=state \\
            connstate=RELATED,ESTABLISHED jump=ACCEPT
        salt '*' nftables.build_rule filter input command=insert position=3 \\
            full=True match=state state=related,established jump=accept

        IPv6:
        salt '*' nftables.build_rule match=state \\
            connstate=related,established jump=accept \\
            family=ipv6
        salt '*' nftables.build_rule filter input command=insert position=3 \\
            full=True match=state state=related,established jump=accept \\
            family=ipv6

    """
    ret = {"comment": "", "rule": "", "result": False}

    if "target" in kwargs:
        kwargs["jump"] = kwargs["target"]
        del kwargs["target"]

    for ignore in list(_STATE_INTERNAL_KEYWORDS) + ["chain", "save", "table"]:
        if ignore in kwargs:
            del kwargs[ignore]

    rule = ""
    proto = ""

    nft_family = _NFTABLES_FAMILIES[family]

    if "if" in kwargs:
        rule += "meta iifname {} ".format(kwargs["if"])
        del kwargs["if"]

    if "of" in kwargs:
        rule += "meta oifname {} ".format(kwargs["of"])
        del kwargs["of"]

    if "proto" in kwargs:
        proto = kwargs["proto"]

    if "state" in kwargs:
        del kwargs["state"]

    if "connstate" in kwargs:
        rule += "ct state {{ {0} }} ".format(kwargs["connstate"])
        del kwargs["connstate"]

    if "icmp-type" in kwargs:
        rule += "icmp type {{ {0} }} ".format(kwargs["icmp-type"])
        del kwargs["icmp-type"]

    if "icmpv6-type" in kwargs:
        rule += "icmpv6 type {{ {0} }} ".format(kwargs["icmpv6-type"])
        del kwargs["icmpv6-type"]

    if "pkttype" in kwargs:
        rule += "meta pkttype {{ {0} }} ".format(kwargs["pkttype"])
        del kwargs["pkttype"]

    if "counter" in kwargs:
        rule += "counter "
        del kwargs["counter"]

    if "saddr" in kwargs or "source" in kwargs:
        rule += "{} saddr {} ".format(
            nft_family, kwargs.get("saddr") or kwargs.get("source")
        )
        if "saddr" in kwargs:
            del kwargs["saddr"]
        if "source" in kwargs:
            del kwargs["source"]

    if "daddr" in kwargs or "destination" in kwargs:
        rule += "{} daddr {} ".format(
            nft_family, kwargs.get("daddr") or kwargs.get("destination")
        )
        if "daddr" in kwargs:
            del kwargs["daddr"]
        if "destination" in kwargs:
            del kwargs["destination"]

    if "dport" in kwargs:
        kwargs["dport"] = str(kwargs["dport"])
        if ":" in kwargs["dport"]:
            kwargs["dport"] = kwargs["dport"].replace(":", "-")
        rule += "dport {{ {0} }} ".format(kwargs["dport"])
        del kwargs["dport"]

    if "sport" in kwargs:
        kwargs["sport"] = str(kwargs["sport"])
        if ":" in kwargs["sport"]:
            kwargs["sport"] = kwargs["sport"].replace(":", "-")
        rule += "sport {{ {0} }} ".format(kwargs["sport"])
        del kwargs["sport"]

    if "dports" in kwargs:
        # nftables reverse sorts the ports from
        # high to low, create rule like this
        # so that the check will work
        _dports = kwargs["dports"].split(",")
        _dports = [int(x) for x in _dports]
        _dports.sort(reverse=True)
        kwargs["dports"] = ", ".join(str(x) for x in _dports)

        rule += "dport {{ {0} }} ".format(kwargs["dports"])
        del kwargs["dports"]

    if "sports" in kwargs:
        # nftables reverse sorts the ports from
        # high to low, create rule like this
        # so that the check will work
        _sports = kwargs["sports"].split(",")
        _sports = [int(x) for x in _sports]
        _sports.sort(reverse=True)
        kwargs["sports"] = ", ".join(str(x) for x in _sports)

        rule += "sport {{ {0} }} ".format(kwargs["sports"])
        del kwargs["sports"]

    # Jumps should appear last, except for any arguments that are passed to
    # jumps, which of course need to follow.
    after_jump = []

    if "jump" in kwargs:
        after_jump.append("{} ".format(kwargs["jump"]))
        del kwargs["jump"]

    if "j" in kwargs:
        after_jump.append("{} ".format(kwargs["j"]))
        del kwargs["j"]

    if "redirect-to" in kwargs or "to-port" in kwargs:
        after_jump.append(
            "redirect to {} ".format(kwargs.get("redirect-to") or kwargs.get("to-port"))
        )
        if "redirect-to" in kwargs:
            del kwargs["redirect-to"]
        if "to-port" in kwargs:
            del kwargs["to-port"]

    if "to-ports" in kwargs:
        after_jump.append("--to-ports {} ".format(kwargs["to-ports"]))
        del kwargs["to-ports"]

    if "to-source" in kwargs:
        after_jump.append("{} ".format(kwargs["to-source"]))
        del kwargs["to-source"]

    if "to-destination" in kwargs:
        after_jump.append("{} ".format(kwargs["to-destination"]))
        del kwargs["to-destination"]

    if "reject-with" in kwargs:
        after_jump.append("reject with {} ".format(kwargs["reject-with"]))
        del kwargs["reject-with"]

    for item in after_jump:
        rule += item

    # Strip trailing spaces off rule
    rule = rule.strip()

    # Insert the protocol prior to dport or sport
    rule = rule.replace("dport", f"{proto} dport")
    rule = rule.replace("sport", f"{proto} sport")

    ret["rule"] = rule

    if full in ["True", "true"]:

        if not table:
            ret["comment"] = "Table needs to be specified"
            return ret

        if not chain:
            ret["comment"] = "Chain needs to be specified"
            return ret

        if not command:
            ret["comment"] = "Command needs to be specified"
            return ret

        if command in ["Insert", "insert", "INSERT"]:
            if position:
                ret["rule"] = "{} insert rule {} {} {} position {} {}".format(
                    _nftables_cmd(), nft_family, table, chain, position, rule
                )
            else:
                ret["rule"] = "{} insert rule {} {} {} {}".format(
                    _nftables_cmd(), nft_family, table, chain, rule
                )
        else:
            ret["rule"] = "{} {} rule {} {} {} {}".format(
                _nftables_cmd(), command, nft_family, table, chain, rule
            )

    if ret["rule"]:
        ret["comment"] = "Successfully built rule"
    ret["result"] = True
    return ret


def get_saved_rules(conf_file=None):
    """
    Return a data structure of the rules in the conf file

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.get_saved_rules

    """
    if _conf() and not conf_file:
        conf_file = _conf()

    with salt.utils.files.fopen(conf_file) as fp_:
        lines = salt.utils.data.decode(fp_.readlines())
    rules = []
    for line in lines:
        tmpline = line.strip()
        if not tmpline:
            continue
        if tmpline.startswith("#"):
            continue
        rules.append(line)
    return rules


def list_tables(family="ipv4"):
    """
    Return a data structure of the current, in-memory tables

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.list_tables

        salt '*' nftables.list_tables family=ipv6

    """
    nft_family = _NFTABLES_FAMILIES[family]
    tables = []
    cmd = "{} --json --numeric --numeric --numeric list tables {}".format(
        _nftables_cmd(), nft_family
    )
    out = __salt__["cmd.run"](cmd, python_shell=False)
    if not out:
        return tables

    try:
        data = json.loads(out)
    except ValueError:
        return tables

    if not data or not data.get("nftables"):
        return tables

    for item in data.get("nftables", []):
        if "metainfo" not in item:
            tables.append(item["table"])
    log.debug(tables)
    return tables


def get_rules(family="ipv4"):
    """
    Return a data structure of the current, in-memory rules

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.get_rules

        salt '*' nftables.get_rules family=ipv6

    """
    tables = list_tables(family)
    nft_family = _NFTABLES_FAMILIES[family]

    rules = []
    for table in tables:
        table_name = table["name"]
        cmd = "{} --numeric --numeric --numeric list table {} {}".format(
            _nftables_cmd(), nft_family, table_name
        )
        out = __salt__["cmd.run"](cmd, python_shell=False)
        rules.append(out)
    return rules


def get_rules_json(family="ipv4"):
    """
    .. versionadded:: 3002

    Return a list of dictionaries comprising the current, in-memory rules

    family
        Networking family, either ipv4 or ipv6

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.get_rules_json

        salt '*' nftables.get_rules_json family=ipv6

    """
    nft_family = _NFTABLES_FAMILIES[family]
    rules = []
    cmd = "{} --numeric --numeric --numeric --json list ruleset {}".format(
        _nftables_cmd(), nft_family
    )
    out = __salt__["cmd.run"](cmd, python_shell=False)
    if not out:
        return rules

    try:
        rules = (json.loads(out))["nftables"]
    except (KeyError, ValueError):
        return rules

    return rules


def save(filename=None, family="ipv4"):
    """
    .. versionchanged:: 3002

    Save the current in-memory rules to disk. On systems where /etc/nftables is
    a directory, a file named salt-all-in-one.nft will be dropped inside by default.
    The main nftables configuration will need to include this file.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.save /etc/nftables
    """
    if _conf() and not filename:
        filename = _conf()

    # Invert the dictionary twice to get unique values only.
    nft_families = {v: k for k, v in _NFTABLES_FAMILIES.items()}
    nft_families = {v: k for k, v in nft_families.items()}

    rules = "#! nft -f\n"

    for family in nft_families:
        out = get_rules(family)
        if out:
            rules += "\n"
        rules = rules + "\n".join(out)
    rules = rules + "\n"

    if __salt__["file.directory_exists"](filename):
        filename = f"{filename}/salt-all-in-one.nft"

    try:
        with salt.utils.files.fopen(filename, "wb") as _fh:
            # Write out any changes
            _fh.write(salt.utils.data.encode(rules))
    except OSError as exc:
        raise CommandExecutionError(f"Problem writing to configuration file: {exc}")
    return rules


def get_rule_handle(table="filter", chain=None, rule=None, family="ipv4"):
    """
    Get the handle for a particular rule

    This function accepts a rule in a standard nftables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.get_rule_handle filter input \\
            rule='tcp dport 22 log accept'

        IPv6:
        salt '*' nftables.get_rule_handle filter input \\
            rule='tcp dport 22 log accept' \\
            family=ipv6
    """
    ret = {"comment": "", "result": False}

    if not chain:
        ret["comment"] = "Chain needs to be specified"
        return ret

    if not rule:
        ret["comment"] = "Rule needs to be specified"
        return ret

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    res = check_chain(table, chain, family=family)
    if not res["result"]:
        return res

    res = check(table, chain, rule, family=family)
    if not res["result"]:
        return res

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = "{} --numeric --numeric --numeric --handle list chain {} {} {}".format(
        _nftables_cmd(), nft_family, table, chain
    )
    out = __salt__["cmd.run"](cmd, python_shell=False)
    rules = re.split("\n+", out)

    pat = re.compile(rf"{rule} # handle (?P<handle>\d+)")
    for r in rules:
        match = pat.search(r)
        if match:
            return {"result": True, "handle": match.group("handle")}
    return {"result": False, "comment": f"Could not find rule {rule}"}


def check(table="filter", chain=None, rule=None, family="ipv4"):
    """
    Check for the existence of a rule in the table and chain

    This function accepts a rule in a standard nftables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.check filter input \\
            rule='tcp dport 22 log accept'

        IPv6:
        salt '*' nftables.check filter input \\
            rule='tcp dport 22 log accept' \\
            family=ipv6
    """
    ret = {"comment": "", "result": False}

    if not chain:
        ret["comment"] = "Chain needs to be specified"
        return ret

    if not rule:
        ret["comment"] = "Rule needs to be specified"
        return ret

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    res = check_chain(table, chain, family=family)
    if not res["result"]:
        return res

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = f"{_nftables_cmd()} --handle list chain {nft_family} {table} {chain}"
    search_rule = f"{rule} #".replace("{ ", "{? ?").replace(" }", " ?}?")
    out = __salt__["cmd.run"](cmd, python_shell=False)
    found = re.search(search_rule, out)

    if not found:
        ret["comment"] = (
            "Rule {} in chain {} in table {} in family {} does not exist".format(
                rule, chain, table, family
            )
        )
    else:
        ret["comment"] = "Rule {} in chain {} in table {} in family {} exists".format(
            rule, chain, table, family
        )
        ret["result"] = True
    return ret


def check_chain(table="filter", chain=None, family="ipv4"):
    """
    .. versionadded:: 2014.7.0

    Check for the existence of a chain in the table

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.check_chain filter input

        IPv6:
        salt '*' nftables.check_chain filter input family=ipv6
    """

    ret = {"comment": "", "result": False}

    if not chain:
        ret["comment"] = "Chain needs to be specified"
        return ret

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = f"{_nftables_cmd()} list table {nft_family} {table}"
    out = __salt__["cmd.run"](cmd, python_shell=False).find(f"chain {chain} {{")

    if out == -1:
        ret["comment"] = "Chain {} in table {} in family {} does not exist".format(
            chain, table, family
        )
    else:
        ret["comment"] = "Chain {} in table {} in family {} exists".format(
            chain, table, family
        )
        ret["result"] = True
    return ret


def check_table(table=None, family="ipv4"):
    """
    Check for the existence of a table

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.check_table nat
    """
    ret = {"comment": "", "result": False}

    if not table:
        ret["comment"] = "Table needs to be specified"
        return ret

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = f"{_nftables_cmd()} list tables {nft_family}"
    out = __salt__["cmd.run"](cmd, python_shell=False).find(
        f"table {nft_family} {table}"
    )

    if out == -1:
        ret["comment"] = f"Table {table} in family {family} does not exist"
    else:
        ret["comment"] = f"Table {table} in family {family} exists"
        ret["result"] = True
    return ret


def new_table(table, family="ipv4"):
    """
    .. versionadded:: 2014.7.0

    Create new custom table.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.new_table filter

        IPv6:
        salt '*' nftables.new_table filter family=ipv6
    """
    ret = {"comment": "", "result": False}

    if not table:
        ret["comment"] = "Table needs to be specified"
        return ret

    res = check_table(table, family=family)
    if res["result"]:
        return res

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = f"{_nftables_cmd()} add table {nft_family} {table}"
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        ret["comment"] = f"Table {table} in family {family} created"
        ret["result"] = True
    else:
        ret["comment"] = "Table {} in family {} could not be created".format(
            table, family
        )
    return ret


def delete_table(table, family="ipv4"):
    """
    .. versionadded:: 2014.7.0

    Create new custom table.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.delete_table filter

        IPv6:
        salt '*' nftables.delete_table filter family=ipv6
    """
    ret = {"comment": "", "result": False}

    if not table:
        ret["comment"] = "Table needs to be specified"
        return ret

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = f"{_nftables_cmd()} delete table {nft_family} {table}"
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        ret["comment"] = f"Table {table} in family {family} deleted"
        ret["result"] = True
    else:
        ret["comment"] = "Table {} in family {} could not be deleted".format(
            table, family
        )
    return ret


def new_chain(
    table="filter", chain=None, table_type=None, hook=None, priority=None, family="ipv4"
):
    """
    .. versionadded:: 2014.7.0

    Create new chain to the specified table.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.new_chain filter input

        salt '*' nftables.new_chain filter input \\
                table_type=filter hook=input priority=0

        salt '*' nftables.new_chain filter foo

        IPv6:
        salt '*' nftables.new_chain filter input family=ipv6

        salt '*' nftables.new_chain filter input \\
                table_type=filter hook=input priority=0 family=ipv6

        salt '*' nftables.new_chain filter foo family=ipv6
    """
    ret = {"comment": "", "result": False}

    if not chain:
        ret["comment"] = "Chain needs to be specified"
        return ret

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    res = check_chain(table, chain, family=family)
    if res["result"]:
        ret["comment"] = "Chain {} in table {} in family {} already exists".format(
            chain, table, family
        )
        return ret

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = f"{_nftables_cmd()} -- add chain {nft_family} {table} {chain}"
    if table_type or hook or priority:
        if table_type and hook and str(priority):
            cmd = r"{0} \{{ type {1} hook {2} priority {3}\; \}}".format(
                cmd, table_type, hook, priority
            )
        else:
            # Specify one, require all
            ret["comment"] = "Table_type, hook, and priority required."
            return ret

    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        ret["comment"] = "Chain {} in table {} in family {} created".format(
            chain, table, family
        )
        ret["result"] = True
    else:
        ret["comment"] = (
            "Chain {} in table {} in family {} could not be created".format(
                chain, table, family
            )
        )
    return ret


def delete_chain(table="filter", chain=None, family="ipv4"):
    """
    .. versionadded:: 2014.7.0

    Delete the chain from the specified table.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.delete_chain filter input

        salt '*' nftables.delete_chain filter foo

        IPv6:
        salt '*' nftables.delete_chain filter input family=ipv6

        salt '*' nftables.delete_chain filter foo family=ipv6
    """
    ret = {"comment": "", "result": False}

    if not chain:
        ret["comment"] = "Chain needs to be specified"
        return ret

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    res = check_chain(table, chain, family=family)
    if not res["result"]:
        return res

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = f"{_nftables_cmd()} delete chain {nft_family} {table} {chain}"
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        ret["comment"] = "Chain {} in table {} in family {} deleted".format(
            chain, table, family
        )
        ret["result"] = True
    else:
        ret["comment"] = (
            "Chain {} in table {} in family {} could not be deleted".format(
                chain, table, family
            )
        )
    return ret


def append(table="filter", chain=None, rule=None, family="ipv4"):
    """
    Append a rule to the specified table & chain.

    This function accepts a rule in a standard nftables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.append filter input \\
            rule='tcp dport 22 log accept'

        IPv6:
        salt '*' nftables.append filter input \\
            rule='tcp dport 22 log accept' \\
            family=ipv6
    """
    ret = {
        "comment": "Failed to append rule {} to chain {} in table {}.".format(
            rule, chain, table
        ),
        "result": False,
    }

    if not chain:
        ret["comment"] = "Chain needs to be specified"
        return ret

    if not rule:
        ret["comment"] = "Rule needs to be specified"
        return ret

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    res = check_chain(table, chain, family=family)
    if not res["result"]:
        return res

    res = check(table, chain, rule, family=family)
    if res["result"]:
        ret["comment"] = (
            "Rule {} chain {} in table {} in family {} already exists".format(
                rule, chain, table, family
            )
        )
        return ret

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = "{} add rule {} {} {} {}".format(
        _nftables_cmd(), nft_family, table, chain, rule
    )
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        ret["result"] = True
        ret["comment"] = 'Added rule "{}" chain {} in table {} in family {}.'.format(
            rule, chain, table, family
        )
    else:
        ret["comment"] = (
            'Failed to add rule "{}" chain {} in table {} in family {}.'.format(
                rule, chain, table, family
            )
        )
    return ret


def insert(table="filter", chain=None, position=None, rule=None, family="ipv4"):
    """
    Insert a rule into the specified table & chain, at the specified position.

    If position is not specified, rule will be inserted in first position.

    This function accepts a rule in a standard nftables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    CLI Examples:

    .. code-block:: bash

        salt '*' nftables.insert filter input \\
            rule='tcp dport 22 log accept'

        salt '*' nftables.insert filter input position=3 \\
            rule='tcp dport 22 log accept'

        IPv6:
        salt '*' nftables.insert filter input \\
            rule='tcp dport 22 log accept' \\
            family=ipv6

        salt '*' nftables.insert filter input position=3 \\
            rule='tcp dport 22 log accept' \\
            family=ipv6
    """
    ret = {
        "comment": f"Failed to insert rule {rule} to table {table}.",
        "result": False,
    }

    if not chain:
        ret["comment"] = "Chain needs to be specified"
        return ret

    if not rule:
        ret["comment"] = "Rule needs to be specified"
        return ret

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    res = check_chain(table, chain, family=family)
    if not res["result"]:
        return res

    res = check(table, chain, rule, family=family)
    if res["result"]:
        ret["comment"] = (
            "Rule {} chain {} in table {} in family {} already exists".format(
                rule, chain, table, family
            )
        )
        return ret

    nft_family = _NFTABLES_FAMILIES[family]
    if position:
        cmd = "{} insert rule {} {} {} position {} {}".format(
            _nftables_cmd(), nft_family, table, chain, position, rule
        )
    else:
        cmd = "{} insert rule {} {} {} {}".format(
            _nftables_cmd(), nft_family, table, chain, rule
        )
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        ret["result"] = True
        ret["comment"] = 'Added rule "{}" chain {} in table {} in family {}.'.format(
            rule, chain, table, family
        )
    else:
        ret["comment"] = (
            'Failed to add rule "{}" chain {} in table {} in family {}.'.format(
                rule, chain, table, family
            )
        )
    return ret


def delete(table, chain=None, position=None, rule=None, family="ipv4"):
    """
    Delete a rule from the specified table & chain, specifying either the rule
        in its entirety, or the rule's position in the chain.

    This function accepts a rule in a standard nftables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    CLI Examples:

    .. code-block:: bash

        salt '*' nftables.delete filter input position=3

        salt '*' nftables.delete filter input \\
            rule='tcp dport 22 log accept'

        IPv6:
        salt '*' nftables.delete filter input position=3 family=ipv6

        salt '*' nftables.delete filter input \\
            rule='tcp dport 22 log accept' \\
            family=ipv6
    """
    ret = {
        "comment": f"Failed to delete rule {rule} in table {table}.",
        "result": False,
    }

    if position and rule:
        ret["comment"] = "Only specify a position or a rule, not both"
        return ret

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    res = check_chain(table, chain, family=family)
    if not res["result"]:
        return res

    res = check(table, chain, rule, family=family)
    if not res["result"]:
        ret["comment"] = (
            "Rule {} chain {} in table {} in family {} does not exist".format(
                rule, chain, table, family
            )
        )
        return ret

    # nftables rules can only be deleted using the handle
    # if we don't have it, find it.
    if not position:
        position = get_rule_handle(table, chain, rule, family)

    nft_family = _NFTABLES_FAMILIES[family]
    cmd = "{} delete rule {} {} {} handle {}".format(
        _nftables_cmd(), nft_family, table, chain, position
    )
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        ret["result"] = True
        ret["comment"] = (
            'Deleted rule "{}" in chain {} in table {} in family {}.'.format(
                rule, chain, table, family
            )
        )
    else:
        ret["comment"] = (
            'Failed to delete rule "{}" in chain {}  table {} in family {}'.format(
                rule, chain, table, family
            )
        )
    return ret


def flush(table="filter", chain="", family="ipv4"):
    """
    Flush the chain in the specified table, flush all chains in the specified
    table if chain is not specified.

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.flush filter

        salt '*' nftables.flush filter input

        IPv6:
        salt '*' nftables.flush filter input family=ipv6
    """
    ret = {
        "comment": "Failed to flush rules from chain {} in table {}.".format(
            chain, table
        ),
        "result": False,
    }

    res = check_table(table, family=family)
    if not res["result"]:
        return res

    nft_family = _NFTABLES_FAMILIES[family]

    if chain:
        res = check_chain(table, chain, family=family)
        if not res["result"]:
            return res
        cmd = "{} flush chain {} {} {}".format(
            _nftables_cmd(), nft_family, table, chain
        )
        comment = f"from chain {chain} in table {table} in family {family}."
    else:
        cmd = f"{_nftables_cmd()} flush table {nft_family} {table}"
        comment = f"from table {table} in family {family}."
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        ret["result"] = True
        ret["comment"] = f"Flushed rules {comment}"
    else:
        ret["comment"] = f"Failed to flush rules {comment}"
    return ret


def get_policy(table="filter", chain=None, family="ipv4"):
    """
    .. versionadded:: 3002

    Return the current policy for the specified table/chain

    table
        Name of the table containing the chain to check

    chain
        Name of the chain to get the policy for

    family
        Networking family, either ipv4 or ipv6

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.get_policy filter input

        IPv6:
        salt '*' nftables.get_policy filter input family=ipv6
    """
    if not chain:
        return "Error: Chain needs to be specified"

    nft_family = _NFTABLES_FAMILIES[family]

    rules = get_rules_json(family=nft_family)

    try:
        for rule in rules["nftables"]:
            if (
                rule.get("chain", {}).get("name") == chain
                and rule.get("chain", {}).get("type") == table
            ):
                return rule["chain"]["policy"]
    except (KeyError, TypeError, ValueError):
        return None


def set_policy(table="filter", chain=None, policy=None, family="ipv4"):
    """
    .. versionadded:: 3002

    Set the current policy for the specified table/chain. This only works on
    chains with an existing base chain.

    table
        Name of the table containing the chain to modify

    chain
        Name of the chain to set the policy for

    policy
        accept or drop

    family
        Networking family, either ipv4 or ipv6

    CLI Example:

    .. code-block:: bash

        salt '*' nftables.set_policy filter input accept

        IPv6:
        salt '*' nftables.set_policy filter input accept family=ipv6
    """
    if not chain:
        return "Error: Chain needs to be specified"
    if not policy:
        return "Error: Policy needs to be specified"

    nft_family = _NFTABLES_FAMILIES[family]

    chain_info = {}
    rules = get_rules_json(family=nft_family)

    if not rules:
        return False

    for rule in rules:
        try:
            if rule["chain"]["table"] == table and rule["chain"]["name"] == chain:
                chain_info = rule["chain"]
                break
        except KeyError:
            continue

    if not chain_info:
        return False

    cmd = f"{_nftables_cmd()} add chain {nft_family} {table} {chain}"

    # We can't infer the base chain parameters. Bail out if they're not present.
    if "type" not in chain_info or "hook" not in chain_info or "prio" not in chain_info:
        return False

    params = "type {} hook {} priority {};".format(
        chain_info["type"], chain_info["hook"], chain_info["prio"]
    )

    cmd = f'{cmd} "{{ {params} policy {policy}; }}"'

    out = __salt__["cmd.run_all"](cmd, python_shell=False)

    return not out["retcode"]



from typing import Union, Dict, Iterable, Optional

Buffer = bytes|bytearray|memoryview

from Cryptodome.Cipher._mode_ecb import EcbMode
from Cryptodome.Cipher._mode_cbc import CbcMode
from Cryptodome.Cipher._mode_cfb import CfbMode
from Cryptodome.Cipher._mode_ofb import OfbMode
from Cryptodome.Cipher._mode_ctr import CtrMode
from Cryptodome.Cipher._mode_openpgp import OpenPgpMode
from Cryptodome.Cipher._mode_eax import EaxMode

ARC2Mode = int

MODE_ECB: ARC2Mode
MODE_CBC: ARC2Mode
MODE_CFB: ARC2Mode
MODE_OFB: ARC2Mode
MODE_CTR: ARC2Mode
MODE_OPENPGP: ARC2Mode
MODE_EAX: ARC2Mode

def new(key: Buffer,
        mode: ARC2Mode,
        iv : Optional[Buffer] = ...,
        IV : Optional[Buffer] = ...,
        nonce : Optional[Buffer] = ...,
        segment_size : int = ...,
        mac_len : int = ...,
        initial_value : Union[int, Buffer] = ...,
        counter : Dict = ...) -> \
        Union[EcbMode, CbcMode, CfbMode, OfbMode, CtrMode, OpenPgpMode]: ...

block_size: int
key_size: Iterable[int]



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

# Cryptography documentation build configuration file, created by
# sphinx-quickstart on Tue Aug  6 19:19:14 2013.
#
# This file is execfile()d with the current directory set to its containing dir
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

try:
    import sphinx_rtd_theme
except ImportError:
    sphinx_rtd_theme = None

try:
    from sphinxcontrib import spelling
except ImportError:
    spelling = None


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("_ext"))

# -- General configuration ----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions  coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "cryptography-docs",
    "sphinx_rtd_theme",
    "sphinx_inline_tabs",
]

if spelling is not None:
    extensions.append("sphinxcontrib.spelling")

# Linkcode resolver
from linkcode_res import linkcode_resolve  # noqa: E402, F401

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

nitpicky = True

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Cryptography"
copyright = "2013-2025, Individual Contributors"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

base_dir = os.path.join(os.path.dirname(__file__), os.pardir)
about = {}
with open(os.path.join(base_dir, "src", "cryptography", "__about__.py")) as f:
    exec(f.read(), about)

version = release = about["__version__"]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The reST default role (used for this markup: `text`) to use for all documents
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

if sphinx_rtd_theme:
    html_theme = "sphinx_rtd_theme"
else:
    html_theme = "default"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Output file base name for HTML help builder.
htmlhelp_basename = "Cryptographydoc"


# -- Options for LaTeX output -------------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual])
latex_documents = [
    (
        "index",
        "Cryptography.tex",
        "Cryptography Documentation",
        "Individual Contributors",
        "manual",
    ),
]

# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        "index",
        "cryptography",
        "Cryptography Documentation",
        ["Individual Contributors"],
        1,
    )
]

# -- Options for Texinfo output -----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "Cryptography",
        "Cryptography Documentation",
        "Individual Contributors",
        "Cryptography",
        "One line description of project.",
        "Miscellaneous",
    ),
]

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

epub_theme = "epub"

# Retry requests in the linkcheck builder so that we're resillient against
# transient network errors.
linkcheck_retries = 10

linkcheck_timeout = 5

linkcheck_ignore = [
    # Insecure renegotiation settings
    r"https://info.isl.ntt.co.jp/crypt/eng/camellia/",
    # Cloudflare returns 403s for all non-browser requests
    r"https://speakerdeck.com",
    r"https://\w+.stackexchange.com",
    r"https://stackoverflow.com",
    r"https://webstore.ansi.org",
    # Kuleuven struggles with the endless forward march of time
    r"https://www.cosic.esat.kuleuven.be",
    # CMU doesn't know how to send intermediates
    r"https://wiki.sei.cmu.edu",
    # GitHub has started heavily rate-limiting causing job failures
    r"https://github.com",
]

autosectionlabel_prefix_document = True



# Copyright 2022-2026 Broadcom.
# SPDX-License-Identifier: Apache-2.0
"""
Download utility class for fetching build dependencies.
"""

from __future__ import annotations

import hashlib
import logging
import os
import pathlib
import subprocess
import sys
from typing import TYPE_CHECKING

from relenv.common import (
    ChecksumValidationError,
    ConfigurationError,
    RelenvException,
    download_url,
    get_download_location,
    runcmd,
)

if TYPE_CHECKING:
    from collections.abc import Callable

# Type alias for path-like objects
PathLike = str | os.PathLike[str]

# Environment flag for CI/CD detection
CICD = "CI" in os.environ

log = logging.getLogger(__name__)


def verify_checksum(file: PathLike, checksum: str | None) -> bool:
    """
    Verify the checksum of a file.

    Supports both SHA-1 (40 hex chars) and SHA-256 (64 hex chars) checksums.
    The hash algorithm is auto-detected based on checksum length.

    :param file: The path to the file to check.
    :type file: str
    :param checksum: The checksum to verify against (SHA-1 or SHA-256)
    :type checksum: str

    :raises RelenvException: If the checksum verification failed

    :return: True if it succeeded, or False if the checksum was None
    :rtype: bool
    """
    if checksum is None:
        log.error("Can't verify checksum because none was given")
        return False

    # Auto-detect hash type based on length
    # SHA-1: 40 hex chars, SHA-256: 64 hex chars
    if len(checksum) == 64:
        hash_algo = hashlib.sha256()
        hash_name = "sha256"
    elif len(checksum) == 40:
        hash_algo = hashlib.sha1()
        hash_name = "sha1"
    else:
        raise ChecksumValidationError(f"Invalid checksum length {len(checksum)}. Expected 40 (SHA-1) or 64 (SHA-256)")

    with open(file, "rb") as fp:
        hash_algo.update(fp.read())
        file_checksum = hash_algo.hexdigest()
        if checksum != file_checksum:
            raise ChecksumValidationError(
                f"{hash_name} checksum verification failed. expected={checksum} found={file_checksum}"
            )
    return True


class Download:
    """
    A utility that holds information about content to be downloaded.

    :param name: The name of the download
    :type name: str
    :param url: The url of the download
    :type url: str
    :param signature: The signature of the download, defaults to None
    :type signature: str
    :param destination: The path to download the file to
    :type destination: str
    :param version: The version of the content to download
    :type version: str
    :param sha1: The sha1 sum of the download
    :type sha1: str

    """

    def __init__(
        self,
        name: str,
        url: str,
        fallback_url: str | None = None,
        signature: str | None = None,
        destination: PathLike = "",
        version: str = "",
        checksum: str | None = None,
    ) -> None:
        self.name = name
        self.url_tpl = url
        self.fallback_url_tpl = fallback_url
        self.signature_tpl = signature
        self._destination: pathlib.Path = pathlib.Path()
        if destination:
            self._destination = pathlib.Path(destination)
        self.version = version
        self.checksum = checksum

    def copy(self) -> Download:
        """Create a copy of this Download instance."""
        return Download(
            self.name,
            self.url_tpl,
            self.fallback_url_tpl,
            self.signature_tpl,
            self.destination,
            self.version,
            self.checksum,
        )

    @property
    def destination(self) -> pathlib.Path:
        """Get the destination directory path."""
        return self._destination

    @destination.setter
    def destination(self, value: PathLike | None) -> None:
        """Set the destination directory path."""
        if value:
            self._destination = pathlib.Path(value)
        else:
            self._destination = pathlib.Path()

    @property
    def url(self) -> str:
        """Get the formatted download URL."""
        return self.url_tpl.format(version=self.version)

    @property
    def fallback_url(self) -> str | None:
        """Get the formatted fallback URL if configured."""
        if self.fallback_url_tpl:
            return self.fallback_url_tpl.format(version=self.version)
        return None

    @property
    def signature_url(self) -> str:
        """Get the formatted signature URL."""
        if self.signature_tpl is None:
            raise ConfigurationError("Signature template not configured")
        return self.signature_tpl.format(version=self.version)

    @property
    def filepath(self) -> pathlib.Path:
        """Get the full file path where the download will be saved."""
        _, name = self.url.rsplit("/", 1)
        return self.destination / name

    @property
    def formatted_url(self) -> str:
        """Get the formatted URL (alias for url property)."""
        return self.url_tpl.format(version=self.version)

    def fetch_file(self, progress_callback: Callable[[int, int], None] | None = None) -> tuple[str, bool]:
        """
        Download the file.

        :param progress_callback: Optional callback(downloaded_bytes, total_bytes)
        :type progress_callback: Optional[Callable[[int, int], None]]
        :return: The path to the downloaded content, and whether it was downloaded.
        :rtype: tuple(str, bool)
        """
        try:
            return (
                download_url(
                    self.url,
                    self.destination,
                    CICD,
                    progress_callback=progress_callback,
                ),
                True,
            )
        except Exception as exc:
            fallback = self.fallback_url
            if fallback:
                print(f"Download failed {self.url} ({exc}); trying fallback url")
                return (
                    download_url(
                        fallback,
                        self.destination,
                        CICD,
                        progress_callback=progress_callback,
                    ),
                    True,
                )
            raise

    def fetch_signature(self, version: str | None = None) -> tuple[str, bool]:
        """
        Download the file signature.

        :return: The path to the downloaded signature.
        :rtype: str
        """
        return download_url(self.signature_url, self.destination, CICD), True

    def exists(self) -> bool:
        """
        True when the artifact already exists on disk.

        :return: True when the artifact already exists on disk
        :rtype: bool
        """
        return self.filepath.exists()

    def valid_hash(self) -> None:
        """Validate the hash of the downloaded file (placeholder method)."""
        pass

    @staticmethod
    def validate_signature(archive: PathLike, signature: PathLike | None) -> bool:
        """
        True when the archive's signature is valid.

        :param archive: The path to the archive to validate
        :type archive: str
        :param signature: The path to the signature to validate against
        :type signature: str

        :return: True if it validated properly, else False
        :rtype: bool
        """
        if signature is None:
            log.error("Can't check signature because none was given")
            return False
        try:
            runcmd(
                ["gpg", "--verify", signature, archive],
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            return True
        except RelenvException as exc:
            log.error("Signature validation failed on %s: %s", archive, exc)
            return False

    @staticmethod
    def validate_checksum(archive: PathLike, checksum: str | None) -> bool:
        """
        True when when the archive matches the sha1 hash.

        :param archive: The path to the archive to validate
        :type archive: str
        :param checksum: The sha1 sum to validate against
        :type checksum: str
        :return: True if the sums matched, else False
        :rtype: bool
        """
        try:
            verify_checksum(archive, checksum)
            return True
        except RelenvException as exc:
            log.error("sha1 validation failed on %s: %s", archive, exc)
            return False

    def __call__(
        self,
        force_download: bool = False,
        show_ui: bool = False,
        exit_on_failure: bool = False,
        progress_callback: Callable[[int, int], None] | None = None,
    ) -> bool:
        """
        Downloads the url and validates the signature and sha1 sum.

        :param progress_callback: Optional callback(downloaded_bytes, total_bytes)
        :type progress_callback: Optional[Callable[[int, int], None]]
        :return: Whether or not validation succeeded
        :rtype: bool
        """
        os.makedirs(self.filepath.parent, exist_ok=True)

        downloaded = False
        if force_download:
            _, downloaded = self.fetch_file(progress_callback)
        else:
            file_is_valid = False
            dest = get_download_location(self.url, self.destination)
            if self.checksum and os.path.exists(dest):
                file_is_valid = self.validate_checksum(dest, self.checksum)
            if file_is_valid:
                log.debug("%s already downloaded, skipping.", self.url)
            else:
                _, downloaded = self.fetch_file(progress_callback)
        valid = True
        if downloaded:
            if self.signature_tpl is not None:
                sig, _ = self.fetch_signature()
                valid_sig = self.validate_signature(self.filepath, sig)
                valid = valid and valid_sig
            if self.checksum is not None:
                valid_checksum = self.validate_checksum(self.filepath, self.checksum)
                valid = valid and valid_checksum

            if not valid:
                log.warning("Checksum did not match %s: %s", self.name, self.checksum)
                if show_ui:
                    sys.stderr.write(f"\nChecksum did not match {self.name}: {self.checksum}\n")
                    sys.stderr.flush()
        if exit_on_failure and not valid:
            sys.exit(1)
        return valid



"""
SaltPeer — bridges the Raft callback surface to the cluster channel transport.

The Raft core (``salt.cluster.consensus.raft.Node``) talks to peers through
the ``Peer`` interface: fire-and-forget RPCs whose *reply* arrives later via
a callback.  ``SaltPeer`` implements that interface by:

  • **Sending**  — serialising the RPC with ``salt.cluster.consensus.rpc`` and
    pushing it to the remote master's ``cluster_pool_port`` via the per-peer
    ``PublishServer`` pusher that already exists in
    ``MasterPubServerChannel._publish_daemon``.

  • **Receiving** — ``handle_pool_publish`` in the channel server calls
    ``RaftDispatcher.dispatch`` for every ``cluster/raft/*`` tag.
    ``RaftDispatcher`` holds a reference to the local ``Node`` and routes
    each inbound message to the correct ``Node`` method, then fires the
    reply callback (which writes the reply back through the sender's pusher).

Asyncio is used for all I/O; the Raft node methods themselves remain
synchronous and callback-oriented.
"""

import asyncio
import logging
import uuid

from salt.cluster.consensus import rpc
from salt.cluster.consensus.raft.node import Peer

log = logging.getLogger(__name__)


async def _publish(pusher, raw):
    """
    Send *raw* over a Raft peer's TCP pusher using a truly-async path.

    Why this exists
    ---------------
    ``salt.transport.tcp.PublishServer.publish`` is declared ``async def``
    but its body is synchronous: it drives a Tornado event loop via
    :class:`salt.utils.asynchronous.SyncWrapper`.  Awaiting it directly
    on a busy asyncio loop blocks the loop on the underlying TCP
    connect/send retry; offloading to ``loop.run_in_executor`` (the
    previous shape of this code) makes the executor thread invoke
    ``SyncWrapper.run_sync`` from outside the loop's thread, which
    races with loop teardown — under CPU contention or fixture
    shutdown the thread schedules late, finds the loop stopped, and
    raises ``RuntimeError: Event loop stopped before Future completed.``
    The Raft RPC is silently dropped, elections never converge, and
    the test eventually fails with "no leader elected" or split-brain.

    Local repro of the bug pre-fix: 7/20 fail under stress-ng on
    debian-12 amd64.  Post-fix: 0/20 expected.

    What this does
    --------------
    Lazily attach a private :class:`salt.transport.tcp._TCPPubServerPublisher`
    to the pusher object on first send and reuse it on subsequent
    sends.  That class exposes a real ``async def send`` that uses the
    underlying Tornado IOStream directly — no SyncWrapper, no executor,
    no cross-thread loop access.

    Why we don't change ``salt.transport.tcp.PublishServer``
    -------------------------------------------------------
    Adding a public ``publish_async`` method to ``PublishServer`` would
    expand salt's transport API surface, which is a salt-wide decision
    requiring buy-in across all transport implementations.  Keeping the
    truly-async client as a private attribute on the pusher (only used
    by our consensus code) confines the change to this branch.

    Test fakes / mocks
    ------------------
    Pushers that aren't real ``PublishServer`` instances (test fakes,
    in-memory mocks, etc.) are detected by class module — their
    ``publish`` is already truly async, so just await it directly.
    ``MagicMock`` auto-creates any attribute on access, so a
    ``hasattr`` probe for ``pull_host`` / ``pull_port`` would falsely
    classify a mock as a real publisher; the module check avoids that.
    """
    module = getattr(type(pusher), "__module__", "") or ""
    if not module.startswith("salt.transport"):
        await pusher.publish(raw)
        return

    client = getattr(pusher, "_consensus_async_client", None)
    if client is None:
        # Lazy import — keeps this module loadable in test environments
        # that monkey-patch out salt.transport.
        from salt.transport.tcp import (  # pylint: disable=import-outside-toplevel
            _TCPPubServerPublisher,
        )

        client = _TCPPubServerPublisher(
            pusher.pull_host, pusher.pull_port, getattr(pusher, "pull_path", None)
        )
        await client.connect()
        # Stash on the pusher so the next send reuses the connection.
        # The pusher's lifetime exceeds the master process; the kernel
        # closes the fd at exit, so explicit teardown isn't required.
        pusher._consensus_async_client = client
    await client.send(raw)


class SaltPeer(Peer):
    """
    A ``Peer`` that sends Raft RPCs over the cluster pool channel.

    :param node_id:       The remote master's node-id — its
                          ``opts["interface"]`` address (matches the
                          entries in ``opts["cluster_peers"]`` and the
                          ``peer_pushers`` keys used everywhere else
                          in the cluster code).
    :param pusher:        The ``PublishServer`` instance already
                          connected to that master's
                          ``cluster_pool_port``.
    :param local_id:      This master's own node-id (used as ``src``
                          in envelopes).
    :param voting:        Whether the peer counts toward quorum.
    :param raft_group_id: Which Raft group this peer belongs to.
                          ``"cluster"`` (default) is the main cluster
                          group; per-ring peers carry the ring name so
                          the dispatcher on the receiving side routes
                          RPCs to the correct local ``Node``.
    """

    def __init__(self, node_id, pusher, local_id, voting=True, raft_group_id="cluster"):
        # Pass None for the node object; we manage node_id directly.
        super().__init__(None, node_id=node_id, voting=voting)
        self._pusher = pusher
        self._local_id = local_id
        self._raft_group_id = raft_group_id

    @property
    def node_id(self):
        return self._node_id

    @property
    def address(self):
        return self._node_id

    # ------------------------------------------------------------------
    # Internal send helper
    # ------------------------------------------------------------------

    async def _send(self, tag, payload, rpc_id=None):
        rpc_id = rpc_id or str(uuid.uuid4())
        raw = rpc.pack(
            tag,
            self._local_id,
            rpc_id,
            payload,
            raft_group_id=self._raft_group_id,
        )
        try:
            await _publish(self._pusher, raw)
        except Exception:  # pylint: disable=broad-except
            log.exception("SaltPeer: failed to send %s to %s", tag, self._node_id)

    def _fire(self, tag, payload, rpc_id=None):
        """Schedule ``_send`` on the running event loop (non-blocking)."""
        try:
            loop = asyncio.get_running_loop()
            loop.create_task(self._send(tag, payload, rpc_id))
        except RuntimeError:
            # No running loop — fall back to a new one (test / bootstrap context).
            asyncio.run(self._send(tag, payload, rpc_id))

    # ------------------------------------------------------------------
    # Peer interface
    # ------------------------------------------------------------------

    def request_vote(self, callback, node_id, term, last_log_term, last_log_index):
        self._fire(
            rpc.REQUEST_VOTE,
            {
                "callback_node": self._local_id,
                "candidate_id": node_id,
                "term": term,
                "last_log_term": last_log_term,
                "last_log_index": last_log_index,
            },
        )

    def pre_request_vote(self, callback, node_id, term, last_log_term, last_log_index):
        self._fire(
            rpc.PRE_REQUEST_VOTE,
            {
                "callback_node": self._local_id,
                "candidate_id": node_id,
                "term": term,
                "last_log_term": last_log_term,
                "last_log_index": last_log_index,
            },
        )

    def append_entries(
        self,
        callback,
        leader_id,
        term,
        prev_log_term,
        prev_log_index,
        leader_commit,
        *entries,
        **kwargs,
    ):
        self._fire(
            rpc.APPEND_ENTRIES,
            {
                "callback_node": self._local_id,
                "leader_id": leader_id,
                "term": term,
                "prev_log_term": prev_log_term,
                "prev_log_index": prev_log_index,
                "leader_commit": leader_commit,
                "entries": [
                    (
                        e
                        if isinstance(e, dict)
                        else e._asdict() if hasattr(e, "_asdict") else list(e)
                    )
                    for e in entries
                ],
                "leader_client_address": kwargs.get("leader_client_address"),
            },
        )

    def install_snapshot(
        self,
        callback,
        leader_id,
        term,
        last_included_index,
        last_included_term,
        data,
        **kwargs,
    ):
        # snapshot data may be bytes — encode as list of ints for msgpack portability
        if isinstance(data, (bytes, bytearray, memoryview)):
            data = list(bytes(data))
        self._fire(
            rpc.INSTALL_SNAPSHOT,
            {
                "callback_node": self._local_id,
                "leader_id": leader_id,
                "term": term,
                "last_included_index": last_included_index,
                "last_included_term": last_included_term,
                "data": data,
            },
        )


class RaftDispatcher:
    """
    Receives decoded ``cluster/raft/*`` messages from ``handle_pool_publish``
    and drives the local Raft ``Node``, then sends the reply back via the
    appropriate pusher.

    One ``RaftDispatcher`` instance lives inside ``MasterPubServerChannel``
    alongside the existing pushers.  When this master hosts multiple
    Raft groups (the main cluster group plus per-ring groups) it owns
    one ``Node`` per group; inbound RPCs carry a ``raft_group_id``
    field that selects the target ``Node``.

    :param node:      Either a single ``Node`` (treated as the
                      ``"cluster"`` group) or a ``dict[str, Node]``
                      mapping group-id to its local ``Node``.  Passing
                      a single Node preserves the pre-multi-ring
                      constructor signature.
    :param local_id:  This master's node-id (``opts["interface"]``) — used as
                      ``src`` in outbound RPC envelopes.
    :param pushers:   Dict mapping peer node-id (interface address) ->
                      ``PublishServer`` pusher.
    """

    def __init__(self, node, local_id, pushers):
        # Normalise to a dict keyed by group-id.  Callers that pass a
        # bare Node (or None) are interpreted as the main cluster
        # group; the dict-of-nodes form is the multi-ring shape.
        if isinstance(node, dict):
            self._nodes = dict(node)
        else:
            self._nodes = {"cluster": node}
        self._local_id = local_id
        # pushers keyed by peer node-id for O(1) lookup
        self._pushers = pushers  # dict[str, PublishServer]

    @property
    def _node(self):
        """
        Backward-compat accessor for callers that reach in for the
        single Node (tests primarily).  Returns the cluster group's
        Node so existing assertions keep working.
        """
        return self._nodes.get("cluster")

    @_node.setter
    def _node(self, node):
        """
        Mutating ``dispatcher._node`` (used by some tests to simulate
        a failed leader losing its Node) maps to the cluster group.
        Setting ``None`` removes the cluster Node entirely so
        :meth:`dispatch` drops inbound RPCs.
        """
        if node is None:
            self._nodes.pop("cluster", None)
        else:
            self._nodes["cluster"] = node

    def register_node(self, raft_group_id, node):
        """
        Add or replace the ``Node`` for *raft_group_id*.

        Used when ``RaftService`` brings up a per-ring Raft group
        after the dispatcher has already been constructed.
        """
        self._nodes[raft_group_id] = node

    def unregister_node(self, raft_group_id):
        """Remove the ``Node`` registered for *raft_group_id*, if any."""
        self._nodes.pop(raft_group_id, None)

    async def _reply(self, dst, tag, payload, raft_group_id="cluster"):
        pusher = self._pushers.get(dst)
        if pusher is None:
            log.warning("RaftDispatcher: no pusher for %s, dropping %s reply", dst, tag)
            return
        raw = rpc.pack(
            tag,
            self._local_id,
            str(uuid.uuid4()),
            payload,
            raft_group_id=raft_group_id,
        )
        try:
            await _publish(pusher, raw)
        except Exception:  # pylint: disable=broad-except
            log.exception("RaftDispatcher: failed to send %s reply to %s", tag, dst)

    async def dispatch(self, tag, src, rpc_id, payload, raft_group_id="cluster"):
        """Route one inbound Raft RPC to the correct Node method."""
        node = self._nodes.get(raft_group_id)
        if node is None:
            log.debug(
                "RaftDispatcher: no node for group %s, dropping %s",
                raft_group_id,
                tag,
            )
            return

        try:
            if tag == rpc.REQUEST_VOTE:
                granted, term, lc = node.request_vote(
                    payload["candidate_id"],
                    payload["term"],
                    last_log_term=payload.get("last_log_term"),
                    last_log_index=payload.get("last_log_index"),
                )
                await self._reply(
                    payload["callback_node"],
                    rpc.REQUEST_VOTE_REPLY,
                    {"granted": granted, "term": term, "voter_id": self._local_id},
                    raft_group_id=raft_group_id,
                )

            elif tag == rpc.PRE_REQUEST_VOTE:
                granted, term, lc = node.pre_request_vote(
                    payload["candidate_id"],
                    payload["term"],
                    last_log_term=payload.get("last_log_term"),
                    last_log_index=payload.get("last_log_index"),
                )
                await self._reply(
                    payload["callback_node"],
                    rpc.PRE_REQUEST_VOTE_REPLY,
                    {"granted": granted, "term": term, "voter_id": self._local_id},
                    raft_group_id=raft_group_id,
                )

            elif tag == rpc.REQUEST_VOTE_REPLY:
                node.request_vote_reply(src, payload["granted"], payload["term"])

            elif tag == rpc.PRE_REQUEST_VOTE_REPLY:
                node.pre_request_vote_reply(src, payload["granted"], payload["term"])

            elif tag == rpc.APPEND_ENTRIES:
                entries = payload.get("entries", [])
                success, term, last_idx, conflict_term, lc = node.handle_append_entries(
                    payload["leader_id"],
                    payload["term"],
                    payload.get("prev_log_term"),
                    payload.get("prev_log_index"),
                    payload.get("leader_commit"),
                    *entries,
                    leader_client_address=payload.get("leader_client_address"),
                )
                sent_log_index = (
                    payload.get("prev_log_index", -1) + len(entries)
                    if payload.get("prev_log_index") is not None
                    else len(entries) - 1
                )
                await self._reply(
                    payload["callback_node"],
                    rpc.APPEND_ENTRIES_REPLY,
                    {
                        "term": payload["term"],
                        "prev_log_term": payload.get("prev_log_term"),
                        "prev_log_index": payload.get("prev_log_index"),
                        "sent_log_index": sent_log_index,
                        "peer_id": self._local_id,
                        "our_term": term,
                        "success": success,
                        "conflict_index": last_idx,
                        "conflict_term": conflict_term,
                    },
                    raft_group_id=raft_group_id,
                )

            elif tag == rpc.APPEND_ENTRIES_REPLY:
                node.append_entries_reply(
                    payload["term"],
                    payload.get("prev_log_term"),
                    payload.get("prev_log_index"),
                    payload.get("sent_log_index"),
                    payload["peer_id"],
                    payload["our_term"],
                    payload["success"],
                    payload.get("conflict_index"),
                    payload.get("conflict_term"),
                )

            elif tag == rpc.INSTALL_SNAPSHOT:
                raw_data = payload.get("data", [])
                if isinstance(raw_data, list):
                    raw_data = bytes(raw_data)
                our_term, lc = node.install_snapshot(
                    payload["leader_id"],
                    payload["term"],
                    payload["last_included_index"],
                    payload["last_included_term"],
                    raw_data,
                )
                await self._reply(
                    payload["callback_node"],
                    rpc.INSTALL_SNAPSHOT_REPLY,
                    {"our_term": our_term, "peer_id": self._local_id},
                    raft_group_id=raft_group_id,
                )

            elif tag == rpc.INSTALL_SNAPSHOT_REPLY:
                node.install_snapshot_reply(payload["peer_id"], payload["our_term"])

            else:
                log.warning("RaftDispatcher: unhandled tag %s", tag)

        except Exception:  # pylint: disable=broad-except
            log.exception("RaftDispatcher: error handling %s from %s", tag, src)



"""
Management of the Salt beacons
==============================

.. versionadded:: 2015.8.0

.. code-block:: yaml

    ps:
      beacon.present:
        - save: True
        - enable: False
        - services:
            salt-master: running
            apache2: stopped

    sh:
      beacon.present: []

    load:
      beacon.present:
        - averages:
            1m:
              - 0.0
              - 2.0
            5m:
              - 0.0
              - 1.5
            15m:
              - 0.1
              - 1.0

    .. versionadded:: 3000

    Beginning in the 3000 release, multiple copies of a beacon can be configured
    using the ``beacon_module`` parameter.

    inotify_infs:
      beacon.present:
        - save: True
        - enable: True
        - files:
           /etc/infs.conf:
             mask:
               - create
               - delete
               - modify
             recurse: True
             auto_add: True
        - interval: 10
        - beacon_module: inotify
        - disable_during_state_run: True

    inotify_ntp:
      beacon.present:
        - save: True
        - enable: True
        - files:
           /etc/ntp.conf:
             mask:
               - create
               - delete
               - modify
             recurse: True
             auto_add: True
        - interval: 10
        - beacon_module: inotify
        - disable_during_state_run: True
"""

import logging

log = logging.getLogger(__name__)


def present(name, save=False, **kwargs):
    """
    Ensure beacon is configured with the included beacon data.

    name
        The name of the beacon to ensure is configured.
    save
        True/False, if True the beacons.conf file be updated too. Default is False.

    Example:

    .. code-block:: yaml

        ps_beacon:
          beacon.present:
            - name: ps
            - save: True
            - enable: False
            - services:
                salt-master: running
                apache2: stopped
    """

    ret = {"name": name, "result": True, "changes": {}, "comment": []}

    current_beacons = __salt__["beacons.list"](return_yaml=False, **kwargs)
    beacon_data = [{k: v} for k, v in kwargs.items()]

    if name in current_beacons:

        if beacon_data == current_beacons[name]:
            ret["comment"].append(f"Job {name} in correct state")
        else:
            if __opts__.get("test"):
                kwargs["test"] = True
                result = __salt__["beacons.modify"](name, beacon_data, **kwargs)
                ret["comment"].append(result["comment"])
                ret["changes"] = result["changes"]
            else:
                result = __salt__["beacons.modify"](name, beacon_data, **kwargs)
                if not result["result"]:
                    ret["result"] = result["result"]
                    ret["comment"] = result["comment"]
                    return ret
                else:
                    if "changes" in result:
                        ret["comment"].append(f"Modifying {name} in beacons")
                        ret["changes"] = result["changes"]
                    else:
                        ret["comment"].append(result["comment"])

    else:
        if __opts__.get("test"):
            kwargs["test"] = True
            result = __salt__["beacons.add"](name, beacon_data, **kwargs)
            ret["comment"].append(result["comment"])
        else:
            result = __salt__["beacons.add"](name, beacon_data, **kwargs)
            if not result["result"]:
                ret["result"] = result["result"]
                ret["comment"] = result["comment"]
                return ret
            else:
                ret["comment"].append(f"Adding {name} to beacons")

    if save:
        if __opts__.get("test"):
            ret["comment"].append(f"Beacon {name} would be saved")
        else:
            __salt__["beacons.save"]()
            ret["comment"].append(f"Beacon {name} saved")

    ret["comment"] = "\n".join(ret["comment"])
    return ret


def absent(name, save=False, **kwargs):
    """
    Ensure beacon is absent.

    name
        The name of the beacon that is ensured absent.
    save
        True/False, if True the beacons.conf file be updated too. Default is False.

    Example:

    .. code-block:: yaml

        remove_beacon:
          beacon.absent:
            - name: ps
            - save: True

    """

    ret = {"name": name, "result": True, "changes": {}, "comment": []}

    current_beacons = __salt__["beacons.list"](return_yaml=False, **kwargs)
    if name in current_beacons:
        if __opts__.get("test"):
            kwargs["test"] = True
            result = __salt__["beacons.delete"](name, **kwargs)
            ret["comment"].append(result["comment"])
        else:
            result = __salt__["beacons.delete"](name, **kwargs)
            if not result["result"]:
                ret["result"] = result["result"]
                ret["comment"] = result["comment"]
                return ret
            else:
                ret["comment"].append(f"Removed {name} from beacons")
    else:
        ret["comment"].append(f"{name} not configured in beacons")

    if save:
        if __opts__.get("test"):
            ret["comment"].append(f"Beacon {name} would be saved")
        else:
            __salt__["beacons.save"]()
            ret["comment"].append(f"Beacon {name} saved")

    ret["comment"] = "\n".join(ret["comment"])
    return ret


def enabled(name, **kwargs):
    """
    Enable a beacon.

    name
        The name of the beacon to enable.

    Example:

    .. code-block:: yaml

        enable_beacon:
          beacon.enabled:
            - name: ps

    """

    ret = {"name": name, "result": True, "changes": {}, "comment": []}

    current_beacons = __salt__["beacons.list"](return_yaml=False, **kwargs)
    if name in current_beacons:
        if __opts__.get("test"):
            kwargs["test"] = True
            result = __salt__["beacons.enable_beacon"](name, **kwargs)
            ret["comment"].append(result["comment"])
        else:
            result = __salt__["beacons.enable_beacon"](name, **kwargs)
            if not result["result"]:
                ret["result"] = result["result"]
                ret["comment"] = result["comment"]
                return ret
            else:
                ret["comment"].append(f"Enabled {name} from beacons")
    else:
        ret["comment"].append(f"{name} not a configured beacon")

    ret["comment"] = "\n".join(ret["comment"])
    return ret


def disabled(name, **kwargs):
    """
    Disable a beacon.

    name
        The name of the beacon to disable.

    Example:

    .. code-block:: yaml

        disable_beacon:
          beacon.disabled:
            - name: psp

    """

    ret = {"name": name, "result": True, "changes": {}, "comment": []}

    current_beacons = __salt__["beacons.list"](return_yaml=False, **kwargs)
    if name in current_beacons:
        if __opts__.get("test"):
            kwargs["test"] = True
            result = __salt__["beacons.disable_beacon"](name, **kwargs)
            ret["comment"].append(result["comment"])
        else:
            result = __salt__["beacons.disable_beacon"](name, **kwargs)
            if not result["result"]:
                ret["result"] = result["result"]
                ret["comment"] = result["comment"]
                return ret
            else:
                ret["comment"].append(f"Disabled beacon {name}.")
    else:
        ret["comment"].append(f"Job {name} is not configured.")

    ret["comment"] = "\n".join(ret["comment"])
    return ret



"""Module holding utility and convenience functions for zmq event monitoring."""

# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.

from __future__ import annotations

import struct
from typing import Awaitable, TypedDict, overload

import zmq
import zmq.asyncio
from zmq.error import _check_version


class _MonitorMessage(TypedDict):
    event: int
    value: int
    endpoint: bytes


def parse_monitor_message(msg: list[bytes]) -> _MonitorMessage:
    """decode zmq_monitor event messages.

    Parameters
    ----------
    msg : list(bytes)
        zmq multipart message that has arrived on a monitor PAIR socket.

        First frame is::

            16 bit event id
            32 bit event value
            no padding

        Second frame is the endpoint as a bytestring

    Returns
    -------
    event : dict
        event description as dict with the keys `event`, `value`, and `endpoint`.
    """
    if len(msg) != 2 or len(msg[0]) != 6:
        raise RuntimeError(f"Invalid event message format: {msg}")
    event_id, value = struct.unpack("=hi", msg[0])
    event: _MonitorMessage = {
        'event': zmq.Event(event_id),
        'value': zmq.Event(value),
        'endpoint': msg[1],
    }
    return event


async def _parse_monitor_msg_async(
    awaitable_msg: Awaitable[list[bytes]],
) -> _MonitorMessage:
    """Like parse_monitor_msg, but awaitable

    Given awaitable message, return awaitable for the parsed monitor message.
    """

    msg = await awaitable_msg
    # 4.0-style event API
    return parse_monitor_message(msg)


@overload
def recv_monitor_message(
    socket: zmq.asyncio.Socket,
    flags: int = 0,
) -> Awaitable[_MonitorMessage]: ...


@overload
def recv_monitor_message(
    socket: zmq.Socket[bytes],
    flags: int = 0,
) -> _MonitorMessage: ...


def recv_monitor_message(
    socket: zmq.Socket,
    flags: int = 0,
) -> _MonitorMessage | Awaitable[_MonitorMessage]:
    """Receive and decode the given raw message from the monitoring socket and return a dict.

    Requires libzmq ≥ 4.0

    The returned dict will have the following entries:
      event : int
        the event id as described in `libzmq.zmq_socket_monitor`
      value : int
        the event value associated with the event, see `libzmq.zmq_socket_monitor`
      endpoint : str
        the affected endpoint

    .. versionchanged:: 23.1
        Support for async sockets added.
        When called with a async socket,
        returns an awaitable for the monitor message.

    Parameters
    ----------
    socket : zmq.Socket
        The PAIR socket (created by other.get_monitor_socket()) on which to recv the message
    flags : int
        standard zmq recv flags

    Returns
    -------
    event : dict
        event description as dict with the keys `event`, `value`, and `endpoint`.
    """

    _check_version((4, 0), 'libzmq event API')
    # will always return a list
    msg = socket.recv_multipart(flags)

    # transparently handle asyncio socket,
    # returns a Future instead of a dict
    if isinstance(msg, Awaitable):
        return _parse_monitor_msg_async(msg)

    # 4.0-style event API
    return parse_monitor_message(msg)


__all__ = ['parse_monitor_message', 'recv_monitor_message']



# Copyright 2022-2026 Broadcom.
# SPDX-License-Identifier: Apache-2.0
"""
Helper for building libraries to install into a relenv environment.
"""

from __future__ import annotations

import json
import logging
import sys
from typing import TYPE_CHECKING, Any, cast

from .common import (
    MACOS_DEVELOPMENT_TARGET,
    PlatformError,
    RelenvEnvironmentError,
    get_toolchain,
    get_triplet,
)

if TYPE_CHECKING:
    import argparse
    import os

log = logging.getLogger()


def setup_parser(
    subparsers: argparse._SubParsersAction[argparse.ArgumentParser],
) -> None:
    """
    Setup the subparser for the ``relenv buildenv`` command.

    :param subparsers: The subparsers object returned from ``add_subparsers``
    :type subparsers: argparse._SubParsersAction
    """
    subparser = subparsers.add_parser("buildenv", description="Relenv build environment")
    subparser.set_defaults(func=main)
    subparser.add_argument(
        "--json",
        default=False,
        action="store_true",
        help=("Output json to stdout instead of export statments"),
    )


def is_relenv() -> bool:
    """
    True when we are in a relenv environment.
    """
    return hasattr(sys, "RELENV")


def buildenv(
    relenv_path: str | os.PathLike[str] | None = None,
) -> dict[str, str]:
    """
    Relenv build environment variable mapping.
    """
    if not relenv_path:
        if not is_relenv():
            raise RelenvEnvironmentError("Not in a relenv environment")
        relenv_path = cast("str | os.PathLike[str]", cast("Any", sys).RELENV)

    if sys.platform != "linux":
        raise PlatformError("buildenv is only supported on Linux")

    toolchain = get_toolchain()
    if not toolchain:
        raise PlatformError("buildenv is only supported on Linux")

    triplet = get_triplet()
    sysroot = f"{toolchain}/{triplet}/sysroot"
    env = {
        "RELENV_BUILDENV": "1",
        "TOOLCHAIN_PATH": f"{toolchain}",
        "TRIPLET": f"{triplet}",
        "RELENV_PATH": f"{relenv_path}",
        "CC": f"{toolchain}/bin/{triplet}-gcc",
        "CXX": f"{toolchain}/bin/{triplet}-g++",
        "CFLAGS": (f"--sysroot={sysroot} -fPIC -I{relenv_path}/include -I{sysroot}/usr/include"),
        "CXXFLAGS": (
            f"--sysroot={sysroot} "
            f"-fPIC "
            f"-I{relenv_path}/include "
            f"-I{sysroot}/usr/include "
            f"-L{relenv_path}/lib -L{sysroot}/lib "
            f"-Wl,-rpath,{relenv_path}/lib"
        ),
        "CPPFLAGS": (f"--sysroot={sysroot} -fPIC -I{relenv_path}/include -I{sysroot}/usr/include"),
        "CMAKE_CFLAGS": (f"--sysroot={sysroot} -fPIC -I{relenv_path}/include -I{sysroot}/usr/include"),
        "LDFLAGS": (f"--sysroot={sysroot} -L{relenv_path}/lib -L{sysroot}/lib -Wl,-rpath,{relenv_path}/lib"),
        "CRATE_CC_NO_DEFAULTS": "1",
        "OPENSSL_DIR": f"{relenv_path}",
        "OPENSSL_INCLUDE_DIR": f"{relenv_path}/include",
        "OPENSSL_LIB_DIR": f"{relenv_path}/lib",
        "PKG_CONFIG_PATH": f"{relenv_path}/lib/pkgconfig",
        "RUSTFLAGS": f"-L {relenv_path}/lib -C link-arg=-Wl,-rpath,{relenv_path}/lib",
    }
    if sys.platform == "darwin":
        env["MACOS_DEVELOPMENT_TARGET"] = MACOS_DEVELOPMENT_TARGET
    return env


def main(args: argparse.Namespace) -> None:
    """
    The entrypoint into the ``relenv buildenv`` command.

    :param args: The args passed to the command
    :type args: argparse.Namespace
    """
    logging.basicConfig(level=logging.INFO)
    if not is_relenv():
        log.error("Not in a relenv environment.")
        sys.exit(1)
    if sys.platform != "linux":
        log.error("buildenv is only supported on Linux.")

    if args.json:
        print(json.dumps(buildenv()))
        sys.exit(0)

    script = ""
    for k, v in buildenv().items():
        script += f'export {k}="{v}"\n'

    print(script)
    sys.exit(0)



"""
Support for Apache

.. note::
    The functions in here are generic functions designed to work with
    all implementations of Apache. Debian-specific functions have been moved into
    deb_apache.py, but will still load under the ``apache`` namespace when a
    Debian-based system is detected.
"""

import io
import logging
import re
import urllib.error
import urllib.request

import salt.utils.data
import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import SaltException

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load the module if apache is installed
    """
    cmd = _detect_os()
    if salt.utils.path.which(cmd):
        return "apache"
    return (
        False,
        "The apache execution module cannot be loaded: apache is not installed.",
    )


def _detect_os():
    """
    Apache commands and paths differ depending on packaging
    """
    # TODO: Add pillar support for the apachectl location
    os_family = __grains__["os_family"]
    if os_family == "RedHat":
        return "apachectl"
    elif os_family == "Debian" or os_family == "Suse":
        return "apache2ctl"
    else:
        return "apachectl"


def version():
    """
    Return server version (``apachectl -v``)

    CLI Example:

    .. code-block:: bash

        salt '*' apache.version
    """
    cmd = f"{_detect_os()} -v"
    out = __salt__["cmd.run"](cmd).splitlines()
    ret = out[0].split(": ")
    return ret[1]


def fullversion():
    """
    Return server version (``apachectl -V``)

    CLI Example:

    .. code-block:: bash

        salt '*' apache.fullversion
    """
    cmd = f"{_detect_os()} -V"
    ret = {}
    ret["compiled_with"] = []
    out = __salt__["cmd.run"](cmd).splitlines()
    # Example
    #  -D APR_HAS_MMAP
    define_re = re.compile(r"^\s+-D\s+")
    for line in out:
        if ": " in line:
            comps = line.split(": ")
            if not comps:
                continue
            ret[comps[0].strip().lower().replace(" ", "_")] = comps[1].strip()
        elif " -D" in line:
            cwith = define_re.sub("", line)
            ret["compiled_with"].append(cwith)
    return ret


def modules():
    """
    Return list of static and shared modules (``apachectl -M``)

    CLI Example:

    .. code-block:: bash

        salt '*' apache.modules
    """
    cmd = f"{_detect_os()} -M"
    ret = {}
    ret["static"] = []
    ret["shared"] = []
    out = __salt__["cmd.run"](cmd).splitlines()
    for line in out:
        comps = line.split()
        if not comps:
            continue
        if "(static)" in line:
            ret["static"].append(comps[0])
        if "(shared)" in line:
            ret["shared"].append(comps[0])
    return ret


def servermods():
    """
    Return list of modules compiled into the server (``apachectl -l``)

    CLI Example:

    .. code-block:: bash

        salt '*' apache.servermods
    """
    cmd = f"{_detect_os()} -l"
    ret = []
    out = __salt__["cmd.run"](cmd).splitlines()
    for line in out:
        if not line:
            continue
        if ".c" in line:
            ret.append(line.strip())
    return ret


def directives():
    """
    Return list of directives together with expected arguments
    and places where the directive is valid (``apachectl -L``)

    CLI Example:

    .. code-block:: bash

        salt '*' apache.directives
    """
    cmd = f"{_detect_os()} -L"
    ret = {}
    out = __salt__["cmd.run"](cmd)
    out = out.replace("\n\t", "\t")
    for line in out.splitlines():
        if not line:
            continue
        comps = line.split("\t")
        desc = "\n".join(comps[1:])
        ret[comps[0]] = desc
    return ret


def vhosts():
    """
    Show the settings as parsed from the config file (currently
    only shows the virtualhost settings) (``apachectl -S``).
    Because each additional virtual host adds to the execution
    time, this command may require a long timeout be specified
    by using ``-t 10``.

    CLI Example:

    .. code-block:: bash

        salt -t 10 '*' apache.vhosts
    """
    cmd = f"{_detect_os()} -S"
    ret = {}
    namevhost = ""
    out = __salt__["cmd.run"](cmd)
    for line in out.splitlines():
        if not line:
            continue
        comps = line.split()
        if "is a NameVirtualHost" in line:
            namevhost = comps[0]
            ret[namevhost] = {}
        else:
            if comps[0] == "default":
                ret[namevhost]["default"] = {}
                ret[namevhost]["default"]["vhost"] = comps[2]
                ret[namevhost]["default"]["conf"] = re.sub(r"\(|\)", "", comps[3])
            if comps[0] == "port":
                ret[namevhost][comps[3]] = {}
                ret[namevhost][comps[3]]["vhost"] = comps[3]
                ret[namevhost][comps[3]]["conf"] = re.sub(r"\(|\)", "", comps[4])
                ret[namevhost][comps[3]]["port"] = comps[1]
    return ret


def signal(signal=None):
    """
    Signals httpd to start, restart, or stop.

    CLI Example:

    .. code-block:: bash

        salt '*' apache.signal restart
    """
    no_extra_args = ("configtest", "status", "fullstatus")
    valid_signals = ("start", "stop", "restart", "graceful", "graceful-stop")

    if signal not in valid_signals and signal not in no_extra_args:
        return
    # Make sure you use the right arguments
    if signal in valid_signals:
        arguments = f" -k {signal}"
    else:
        arguments = f" {signal}"
    cmd = _detect_os() + arguments
    out = __salt__["cmd.run_all"](cmd)

    # A non-zero return code means fail
    if out["retcode"] and out["stderr"]:
        ret = out["stderr"].strip()
    # 'apachectl configtest' returns 'Syntax OK' to stderr
    elif out["stderr"]:
        ret = out["stderr"].strip()
    elif out["stdout"]:
        ret = out["stdout"].strip()
    # No output for something like: apachectl graceful
    else:
        ret = f'Command: "{cmd}" completed successfully!'
    return ret


def useradd(pwfile, user, password, opts=""):
    """
    Add HTTP user using the ``htpasswd`` command. If the ``htpasswd`` file does not
    exist, it will be created. Valid options that can be passed are:

    .. code-block:: text

        n  Don't update file; display results on stdout.
        m  Force MD5 hashing of the password (default).
        d  Force CRYPT(3) hashing of the password.
        p  Do not hash the password (plaintext).
        s  Force SHA1 hashing of the password.

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.useradd /etc/httpd/htpasswd larry badpassword
        salt '*' apache.useradd /etc/httpd/htpasswd larry badpass opts=ns
    """
    return __salt__["webutil.useradd"](pwfile, user, password, opts)


def userdel(pwfile, user):
    """
    Delete HTTP user from the specified ``htpasswd`` file.

    CLI Example:

    .. code-block:: bash

        salt '*' apache.userdel /etc/httpd/htpasswd larry
    """
    return __salt__["webutil.userdel"](pwfile, user)


def server_status(profile="default"):
    """
    Get Information from the Apache server-status handler

    .. note::

        The server-status handler is disabled by default.
        In order for this function to work it needs to be enabled.
        See http://httpd.apache.org/docs/2.2/mod/mod_status.html

    The following configuration needs to exists in pillar/grains.
    Each entry nested in ``apache.server-status`` is a profile of a vhost/server.
    This would give support for multiple apache servers/vhosts.

    .. code-block:: yaml

        apache.server-status:
          default:
            url: http://localhost/server-status
            user: someuser
            pass: password
            realm: 'authentication realm for digest passwords'
            timeout: 5

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.server_status
        salt '*' apache.server_status other-profile
    """
    ret = {
        "Scoreboard": {
            "_": 0,
            "S": 0,
            "R": 0,
            "W": 0,
            "K": 0,
            "D": 0,
            "C": 0,
            "L": 0,
            "G": 0,
            "I": 0,
            ".": 0,
        },
    }

    # Get configuration from pillar
    url = __salt__["config.get"](
        f"apache.server-status:{profile}:url", "http://localhost/server-status"
    )
    user = __salt__["config.get"](f"apache.server-status:{profile}:user", "")
    passwd = __salt__["config.get"](f"apache.server-status:{profile}:pass", "")
    realm = __salt__["config.get"](f"apache.server-status:{profile}:realm", "")
    timeout = __salt__["config.get"](f"apache.server-status:{profile}:timeout", 5)

    # create authentication handler if configuration exists
    if user and passwd:
        basic = urllib.request.HTTPBasicAuthHandler()
        basic.add_password(realm=realm, uri=url, user=user, passwd=passwd)
        digest = urllib.request.HTTPDigestAuthHandler()
        digest.add_password(realm=realm, uri=url, user=user, passwd=passwd)
        urllib.request.install_opener(urllib.request.build_opener(basic, digest))

    # get http data
    url += "?auto"
    try:
        response = urllib.request.urlopen(url, timeout=timeout).read().splitlines()
    except urllib.error.URLError:
        return "error"

    # parse the data
    for line in response:
        splt = line.split(":", 1)
        splt[0] = splt[0].strip()
        splt[1] = splt[1].strip()

        if splt[0] == "Scoreboard":
            for c in splt[1]:
                ret["Scoreboard"][c] += 1
        else:
            if splt[1].isdigit():
                ret[splt[0]] = int(splt[1])
            else:
                ret[splt[0]] = float(splt[1])

    # return the good stuff
    return ret


def _parse_config(conf, slot=None):
    """
    Recursively goes through config structure and builds final Apache configuration

    :param conf: defined config structure
    :param slot: name of section container if needed
    """
    ret = io.StringIO()
    if isinstance(conf, str):
        if slot:
            print(f"{slot} {conf}", file=ret, end="")
        else:
            print(f"{conf}", file=ret, end="")
    elif isinstance(conf, list):
        is_section = False
        for item in conf:
            if "this" in item:
                is_section = True
                slot_this = str(item["this"])
        if is_section:
            print(f"<{slot} {slot_this}>", file=ret)
            for item in conf:
                for key, val in item.items():
                    if key != "this":
                        print(_parse_config(val, str(key)), file=ret)
            print(f"</{slot}>", file=ret)
        else:
            for value in conf:
                print(_parse_config(value, str(slot)), file=ret)
    elif isinstance(conf, dict):
        try:
            print("<{} {}>".format(slot, conf["this"]), file=ret)
        except KeyError:
            raise SaltException(
                'Apache section container "<{}>" expects attribute. '
                'Specify it using key "this".'.format(slot)
            )
        for key, value in conf.items():
            if key != "this":
                if isinstance(value, str):
                    print(f"{key} {value}", file=ret)
                elif isinstance(value, list):
                    print(_parse_config(value, key), file=ret)
                elif isinstance(value, dict):
                    print(_parse_config(value, key), file=ret)
        print(f"</{slot}>", file=ret)

    ret.seek(0)
    return ret.read()


def config(name, config, edit=True):
    """
    Create VirtualHost configuration files

    name
        File for the virtual host
    config
        VirtualHost configurations

    .. note::

        This function is not meant to be used from the command line.
        Config is meant to be an ordered dict of all of the apache configs.

    CLI Example:

    .. code-block:: bash

        salt '*' apache.config /etc/httpd/conf.d/ports.conf config="[{'Listen': '22'}]"
    """

    configs = []
    for entry in config:
        key = next(iter(entry.keys()))
        configs.append(_parse_config(entry[key], key))

    # Python auto-correct line endings
    configstext = "\n".join(salt.utils.data.decode(configs))
    if edit:
        with salt.utils.files.fopen(name, "w") as configfile:
            configfile.write("# This file is managed by Salt.\n")
            configfile.write(salt.utils.stringutils.to_str(configstext))
    return configstext



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Mapping, Sequence

# This is the implementation of the "Any" type as specified by the specifications of OpenTelemetry data model for logs.
# For more details, refer to the OTel specification:
# https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/logs/data-model.md#type-any
AnyValue = (
    str
    | bool
    | int
    | float
    | bytes
    | Sequence["AnyValue"]
    | Mapping[str, "AnyValue"]
    | None
)

AttributeValue = (
    str
    | bool
    | int
    | float
    | Sequence[str]
    | Sequence[bool]
    | Sequence[int]
    | Sequence[float]
)
Attributes = Mapping[str, AttributeValue] | None
AttributesAsKey = tuple[
    tuple[
        str,
        str
        | bool
        | int
        | float
        | tuple[str | None, ...]
        | tuple[bool | None, ...]
        | tuple[int | None, ...]
        | tuple[float | None, ...],
    ],
    ...,
]

_ExtendedAttributes = Mapping[str, "AnyValue"]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper

from google.api import launch_stage_pb2 as _launch_stage_pb2

DESCRIPTOR: _descriptor.FileDescriptor

class ClientLibraryOrganization(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CLIENT_LIBRARY_ORGANIZATION_UNSPECIFIED: _ClassVar[ClientLibraryOrganization]
    CLOUD: _ClassVar[ClientLibraryOrganization]
    ADS: _ClassVar[ClientLibraryOrganization]
    PHOTOS: _ClassVar[ClientLibraryOrganization]
    STREET_VIEW: _ClassVar[ClientLibraryOrganization]
    SHOPPING: _ClassVar[ClientLibraryOrganization]
    GEO: _ClassVar[ClientLibraryOrganization]
    GENERATIVE_AI: _ClassVar[ClientLibraryOrganization]

class ClientLibraryDestination(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CLIENT_LIBRARY_DESTINATION_UNSPECIFIED: _ClassVar[ClientLibraryDestination]
    GITHUB: _ClassVar[ClientLibraryDestination]
    PACKAGE_MANAGER: _ClassVar[ClientLibraryDestination]

class FlowControlLimitExceededBehaviorProto(
    int, metaclass=_enum_type_wrapper.EnumTypeWrapper
):
    __slots__ = ()
    UNSET_BEHAVIOR: _ClassVar[FlowControlLimitExceededBehaviorProto]
    THROW_EXCEPTION: _ClassVar[FlowControlLimitExceededBehaviorProto]
    BLOCK: _ClassVar[FlowControlLimitExceededBehaviorProto]
    IGNORE: _ClassVar[FlowControlLimitExceededBehaviorProto]

CLIENT_LIBRARY_ORGANIZATION_UNSPECIFIED: ClientLibraryOrganization
CLOUD: ClientLibraryOrganization
ADS: ClientLibraryOrganization
PHOTOS: ClientLibraryOrganization
STREET_VIEW: ClientLibraryOrganization
SHOPPING: ClientLibraryOrganization
GEO: ClientLibraryOrganization
GENERATIVE_AI: ClientLibraryOrganization
CLIENT_LIBRARY_DESTINATION_UNSPECIFIED: ClientLibraryDestination
GITHUB: ClientLibraryDestination
PACKAGE_MANAGER: ClientLibraryDestination
UNSET_BEHAVIOR: FlowControlLimitExceededBehaviorProto
THROW_EXCEPTION: FlowControlLimitExceededBehaviorProto
BLOCK: FlowControlLimitExceededBehaviorProto
IGNORE: FlowControlLimitExceededBehaviorProto
METHOD_SIGNATURE_FIELD_NUMBER: _ClassVar[int]
method_signature: _descriptor.FieldDescriptor
DEFAULT_HOST_FIELD_NUMBER: _ClassVar[int]
default_host: _descriptor.FieldDescriptor
OAUTH_SCOPES_FIELD_NUMBER: _ClassVar[int]
oauth_scopes: _descriptor.FieldDescriptor
API_VERSION_FIELD_NUMBER: _ClassVar[int]
api_version: _descriptor.FieldDescriptor

class CommonLanguageSettings(_message.Message):
    __slots__ = ("reference_docs_uri", "destinations", "selective_gapic_generation")
    REFERENCE_DOCS_URI_FIELD_NUMBER: _ClassVar[int]
    DESTINATIONS_FIELD_NUMBER: _ClassVar[int]
    SELECTIVE_GAPIC_GENERATION_FIELD_NUMBER: _ClassVar[int]
    reference_docs_uri: str
    destinations: _containers.RepeatedScalarFieldContainer[ClientLibraryDestination]
    selective_gapic_generation: SelectiveGapicGeneration
    def __init__(
        self,
        reference_docs_uri: _Optional[str] = ...,
        destinations: _Optional[_Iterable[_Union[ClientLibraryDestination, str]]] = ...,
        selective_gapic_generation: _Optional[
            _Union[SelectiveGapicGeneration, _Mapping]
        ] = ...,
    ) -> None: ...

class ClientLibrarySettings(_message.Message):
    __slots__ = (
        "version",
        "launch_stage",
        "rest_numeric_enums",
        "java_settings",
        "cpp_settings",
        "php_settings",
        "python_settings",
        "node_settings",
        "dotnet_settings",
        "ruby_settings",
        "go_settings",
    )
    VERSION_FIELD_NUMBER: _ClassVar[int]
    LAUNCH_STAGE_FIELD_NUMBER: _ClassVar[int]
    REST_NUMERIC_ENUMS_FIELD_NUMBER: _ClassVar[int]
    JAVA_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    CPP_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    PHP_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    PYTHON_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    NODE_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    DOTNET_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    RUBY_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    GO_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    version: str
    launch_stage: _launch_stage_pb2.LaunchStage
    rest_numeric_enums: bool
    java_settings: JavaSettings
    cpp_settings: CppSettings
    php_settings: PhpSettings
    python_settings: PythonSettings
    node_settings: NodeSettings
    dotnet_settings: DotnetSettings
    ruby_settings: RubySettings
    go_settings: GoSettings
    def __init__(
        self,
        version: _Optional[str] = ...,
        launch_stage: _Optional[_Union[_launch_stage_pb2.LaunchStage, str]] = ...,
        rest_numeric_enums: bool = ...,
        java_settings: _Optional[_Union[JavaSettings, _Mapping]] = ...,
        cpp_settings: _Optional[_Union[CppSettings, _Mapping]] = ...,
        php_settings: _Optional[_Union[PhpSettings, _Mapping]] = ...,
        python_settings: _Optional[_Union[PythonSettings, _Mapping]] = ...,
        node_settings: _Optional[_Union[NodeSettings, _Mapping]] = ...,
        dotnet_settings: _Optional[_Union[DotnetSettings, _Mapping]] = ...,
        ruby_settings: _Optional[_Union[RubySettings, _Mapping]] = ...,
        go_settings: _Optional[_Union[GoSettings, _Mapping]] = ...,
    ) -> None: ...

class Publishing(_message.Message):
    __slots__ = (
        "method_settings",
        "new_issue_uri",
        "documentation_uri",
        "api_short_name",
        "github_label",
        "codeowner_github_teams",
        "doc_tag_prefix",
        "organization",
        "library_settings",
        "proto_reference_documentation_uri",
        "rest_reference_documentation_uri",
    )
    METHOD_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    NEW_ISSUE_URI_FIELD_NUMBER: _ClassVar[int]
    DOCUMENTATION_URI_FIELD_NUMBER: _ClassVar[int]
    API_SHORT_NAME_FIELD_NUMBER: _ClassVar[int]
    GITHUB_LABEL_FIELD_NUMBER: _ClassVar[int]
    CODEOWNER_GITHUB_TEAMS_FIELD_NUMBER: _ClassVar[int]
    DOC_TAG_PREFIX_FIELD_NUMBER: _ClassVar[int]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    LIBRARY_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    PROTO_REFERENCE_DOCUMENTATION_URI_FIELD_NUMBER: _ClassVar[int]
    REST_REFERENCE_DOCUMENTATION_URI_FIELD_NUMBER: _ClassVar[int]
    method_settings: _containers.RepeatedCompositeFieldContainer[MethodSettings]
    new_issue_uri: str
    documentation_uri: str
    api_short_name: str
    github_label: str
    codeowner_github_teams: _containers.RepeatedScalarFieldContainer[str]
    doc_tag_prefix: str
    organization: ClientLibraryOrganization
    library_settings: _containers.RepeatedCompositeFieldContainer[ClientLibrarySettings]
    proto_reference_documentation_uri: str
    rest_reference_documentation_uri: str
    def __init__(
        self,
        method_settings: _Optional[_Iterable[_Union[MethodSettings, _Mapping]]] = ...,
        new_issue_uri: _Optional[str] = ...,
        documentation_uri: _Optional[str] = ...,
        api_short_name: _Optional[str] = ...,
        github_label: _Optional[str] = ...,
        codeowner_github_teams: _Optional[_Iterable[str]] = ...,
        doc_tag_prefix: _Optional[str] = ...,
        organization: _Optional[_Union[ClientLibraryOrganization, str]] = ...,
        library_settings: _Optional[
            _Iterable[_Union[ClientLibrarySettings, _Mapping]]
        ] = ...,
        proto_reference_documentation_uri: _Optional[str] = ...,
        rest_reference_documentation_uri: _Optional[str] = ...,
    ) -> None: ...

class JavaSettings(_message.Message):
    __slots__ = ("library_package", "service_class_names", "common")
    class ServiceClassNamesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    LIBRARY_PACKAGE_FIELD_NUMBER: _ClassVar[int]
    SERVICE_CLASS_NAMES_FIELD_NUMBER: _ClassVar[int]
    COMMON_FIELD_NUMBER: _ClassVar[int]
    library_package: str
    service_class_names: _containers.ScalarMap[str, str]
    common: CommonLanguageSettings
    def __init__(
        self,
        library_package: _Optional[str] = ...,
        service_class_names: _Optional[_Mapping[str, str]] = ...,
        common: _Optional[_Union[CommonLanguageSettings, _Mapping]] = ...,
    ) -> None: ...

class CppSettings(_message.Message):
    __slots__ = ("common",)
    COMMON_FIELD_NUMBER: _ClassVar[int]
    common: CommonLanguageSettings
    def __init__(
        self, common: _Optional[_Union[CommonLanguageSettings, _Mapping]] = ...
    ) -> None: ...

class PhpSettings(_message.Message):
    __slots__ = ("common", "library_package")
    COMMON_FIELD_NUMBER: _ClassVar[int]
    LIBRARY_PACKAGE_FIELD_NUMBER: _ClassVar[int]
    common: CommonLanguageSettings
    library_package: str
    def __init__(
        self,
        common: _Optional[_Union[CommonLanguageSettings, _Mapping]] = ...,
        library_package: _Optional[str] = ...,
    ) -> None: ...

class PythonSettings(_message.Message):
    __slots__ = ("common", "experimental_features")
    class ExperimentalFeatures(_message.Message):
        __slots__ = (
            "rest_async_io_enabled",
            "protobuf_pythonic_types_enabled",
            "unversioned_package_disabled",
        )
        REST_ASYNC_IO_ENABLED_FIELD_NUMBER: _ClassVar[int]
        PROTOBUF_PYTHONIC_TYPES_ENABLED_FIELD_NUMBER: _ClassVar[int]
        UNVERSIONED_PACKAGE_DISABLED_FIELD_NUMBER: _ClassVar[int]
        rest_async_io_enabled: bool
        protobuf_pythonic_types_enabled: bool
        unversioned_package_disabled: bool
        def __init__(
            self,
            rest_async_io_enabled: bool = ...,
            protobuf_pythonic_types_enabled: bool = ...,
            unversioned_package_disabled: bool = ...,
        ) -> None: ...

    COMMON_FIELD_NUMBER: _ClassVar[int]
    EXPERIMENTAL_FEATURES_FIELD_NUMBER: _ClassVar[int]
    common: CommonLanguageSettings
    experimental_features: PythonSettings.ExperimentalFeatures
    def __init__(
        self,
        common: _Optional[_Union[CommonLanguageSettings, _Mapping]] = ...,
        experimental_features: _Optional[
            _Union[PythonSettings.ExperimentalFeatures, _Mapping]
        ] = ...,
    ) -> None: ...

class NodeSettings(_message.Message):
    __slots__ = ("common",)
    COMMON_FIELD_NUMBER: _ClassVar[int]
    common: CommonLanguageSettings
    def __init__(
        self, common: _Optional[_Union[CommonLanguageSettings, _Mapping]] = ...
    ) -> None: ...

class DotnetSettings(_message.Message):
    __slots__ = (
        "common",
        "renamed_services",
        "renamed_resources",
        "ignored_resources",
        "forced_namespace_aliases",
        "handwritten_signatures",
    )
    class RenamedServicesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    class RenamedResourcesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    COMMON_FIELD_NUMBER: _ClassVar[int]
    RENAMED_SERVICES_FIELD_NUMBER: _ClassVar[int]
    RENAMED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    IGNORED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    FORCED_NAMESPACE_ALIASES_FIELD_NUMBER: _ClassVar[int]
    HANDWRITTEN_SIGNATURES_FIELD_NUMBER: _ClassVar[int]
    common: CommonLanguageSettings
    renamed_services: _containers.ScalarMap[str, str]
    renamed_resources: _containers.ScalarMap[str, str]
    ignored_resources: _containers.RepeatedScalarFieldContainer[str]
    forced_namespace_aliases: _containers.RepeatedScalarFieldContainer[str]
    handwritten_signatures: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        common: _Optional[_Union[CommonLanguageSettings, _Mapping]] = ...,
        renamed_services: _Optional[_Mapping[str, str]] = ...,
        renamed_resources: _Optional[_Mapping[str, str]] = ...,
        ignored_resources: _Optional[_Iterable[str]] = ...,
        forced_namespace_aliases: _Optional[_Iterable[str]] = ...,
        handwritten_signatures: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class RubySettings(_message.Message):
    __slots__ = ("common",)
    COMMON_FIELD_NUMBER: _ClassVar[int]
    common: CommonLanguageSettings
    def __init__(
        self, common: _Optional[_Union[CommonLanguageSettings, _Mapping]] = ...
    ) -> None: ...

class GoSettings(_message.Message):
    __slots__ = ("common", "renamed_services")
    class RenamedServicesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    COMMON_FIELD_NUMBER: _ClassVar[int]
    RENAMED_SERVICES_FIELD_NUMBER: _ClassVar[int]
    common: CommonLanguageSettings
    renamed_services: _containers.ScalarMap[str, str]
    def __init__(
        self,
        common: _Optional[_Union[CommonLanguageSettings, _Mapping]] = ...,
        renamed_services: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class MethodSettings(_message.Message):
    __slots__ = ("selector", "long_running", "auto_populated_fields", "batching")
    class LongRunning(_message.Message):
        __slots__ = (
            "initial_poll_delay",
            "poll_delay_multiplier",
            "max_poll_delay",
            "total_poll_timeout",
        )
        INITIAL_POLL_DELAY_FIELD_NUMBER: _ClassVar[int]
        POLL_DELAY_MULTIPLIER_FIELD_NUMBER: _ClassVar[int]
        MAX_POLL_DELAY_FIELD_NUMBER: _ClassVar[int]
        TOTAL_POLL_TIMEOUT_FIELD_NUMBER: _ClassVar[int]
        initial_poll_delay: _duration_pb2.Duration
        poll_delay_multiplier: float
        max_poll_delay: _duration_pb2.Duration
        total_poll_timeout: _duration_pb2.Duration
        def __init__(
            self,
            initial_poll_delay: _Optional[
                _Union[_duration_pb2.Duration, _Mapping]
            ] = ...,
            poll_delay_multiplier: _Optional[float] = ...,
            max_poll_delay: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
            total_poll_timeout: _Optional[
                _Union[_duration_pb2.Duration, _Mapping]
            ] = ...,
        ) -> None: ...

    SELECTOR_FIELD_NUMBER: _ClassVar[int]
    LONG_RUNNING_FIELD_NUMBER: _ClassVar[int]
    AUTO_POPULATED_FIELDS_FIELD_NUMBER: _ClassVar[int]
    BATCHING_FIELD_NUMBER: _ClassVar[int]
    selector: str
    long_running: MethodSettings.LongRunning
    auto_populated_fields: _containers.RepeatedScalarFieldContainer[str]
    batching: BatchingConfigProto
    def __init__(
        self,
        selector: _Optional[str] = ...,
        long_running: _Optional[_Union[MethodSettings.LongRunning, _Mapping]] = ...,
        auto_populated_fields: _Optional[_Iterable[str]] = ...,
        batching: _Optional[_Union[BatchingConfigProto, _Mapping]] = ...,
    ) -> None: ...

class SelectiveGapicGeneration(_message.Message):
    __slots__ = ("methods", "generate_omitted_as_internal")
    METHODS_FIELD_NUMBER: _ClassVar[int]
    GENERATE_OMITTED_AS_INTERNAL_FIELD_NUMBER: _ClassVar[int]
    methods: _containers.RepeatedScalarFieldContainer[str]
    generate_omitted_as_internal: bool
    def __init__(
        self,
        methods: _Optional[_Iterable[str]] = ...,
        generate_omitted_as_internal: bool = ...,
    ) -> None: ...

class BatchingConfigProto(_message.Message):
    __slots__ = ("thresholds", "batch_descriptor")
    THRESHOLDS_FIELD_NUMBER: _ClassVar[int]
    BATCH_DESCRIPTOR_FIELD_NUMBER: _ClassVar[int]
    thresholds: BatchingSettingsProto
    batch_descriptor: BatchingDescriptorProto
    def __init__(
        self,
        thresholds: _Optional[_Union[BatchingSettingsProto, _Mapping]] = ...,
        batch_descriptor: _Optional[_Union[BatchingDescriptorProto, _Mapping]] = ...,
    ) -> None: ...

class BatchingSettingsProto(_message.Message):
    __slots__ = (
        "element_count_threshold",
        "request_byte_threshold",
        "delay_threshold",
        "element_count_limit",
        "request_byte_limit",
        "flow_control_element_limit",
        "flow_control_byte_limit",
        "flow_control_limit_exceeded_behavior",
    )
    ELEMENT_COUNT_THRESHOLD_FIELD_NUMBER: _ClassVar[int]
    REQUEST_BYTE_THRESHOLD_FIELD_NUMBER: _ClassVar[int]
    DELAY_THRESHOLD_FIELD_NUMBER: _ClassVar[int]
    ELEMENT_COUNT_LIMIT_FIELD_NUMBER: _ClassVar[int]
    REQUEST_BYTE_LIMIT_FIELD_NUMBER: _ClassVar[int]
    FLOW_CONTROL_ELEMENT_LIMIT_FIELD_NUMBER: _ClassVar[int]
    FLOW_CONTROL_BYTE_LIMIT_FIELD_NUMBER: _ClassVar[int]
    FLOW_CONTROL_LIMIT_EXCEEDED_BEHAVIOR_FIELD_NUMBER: _ClassVar[int]
    element_count_threshold: int
    request_byte_threshold: int
    delay_threshold: _duration_pb2.Duration
    element_count_limit: int
    request_byte_limit: int
    flow_control_element_limit: int
    flow_control_byte_limit: int
    flow_control_limit_exceeded_behavior: FlowControlLimitExceededBehaviorProto
    def __init__(
        self,
        element_count_threshold: _Optional[int] = ...,
        request_byte_threshold: _Optional[int] = ...,
        delay_threshold: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        element_count_limit: _Optional[int] = ...,
        request_byte_limit: _Optional[int] = ...,
        flow_control_element_limit: _Optional[int] = ...,
        flow_control_byte_limit: _Optional[int] = ...,
        flow_control_limit_exceeded_behavior: _Optional[
            _Union[FlowControlLimitExceededBehaviorProto, str]
        ] = ...,
    ) -> None: ...

class BatchingDescriptorProto(_message.Message):
    __slots__ = ("batched_field", "discriminator_fields", "subresponse_field")
    BATCHED_FIELD_FIELD_NUMBER: _ClassVar[int]
    DISCRIMINATOR_FIELDS_FIELD_NUMBER: _ClassVar[int]
    SUBRESPONSE_FIELD_FIELD_NUMBER: _ClassVar[int]
    batched_field: str
    discriminator_fields: _containers.RepeatedScalarFieldContainer[str]
    subresponse_field: str
    def __init__(
        self,
        batched_field: _Optional[str] = ...,
        discriminator_fields: _Optional[_Iterable[str]] = ...,
        subresponse_field: _Optional[str] = ...,
    ) -> None: ...



"""
This module allows you to manage extended attributes on files or directories

.. code-block:: bash

    salt '*' xattr.list /path/to/file
"""

import logging

import salt.utils.args
import salt.utils.mac_utils
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

__virtualname__ = "xattr"
__func_alias__ = {
    "list_": "list",
}


def __virtual__():
    """
    Only work on Mac OS
    """
    if __grains__["os"] in ["MacOS", "Darwin"]:
        return __virtualname__
    return (False, "Only available on Mac OS systems")


def list_(path, **kwargs):
    """
    List all of the extended attributes on the given file/directory

    :param str path: The file(s) to get attributes from

    :param bool hex: Return the values with forced hexadecimal values

    :return: A dictionary containing extended attributes and values for the
        given file
    :rtype: dict

    :raises: CommandExecutionError on file not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' xattr.list /path/to/file
        salt '*' xattr.list /path/to/file hex=True
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    hex_ = kwargs.pop("hex", False)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    cmd = ["xattr", path]
    try:
        ret = salt.utils.mac_utils.execute_return_result(cmd)
    except CommandExecutionError as exc:
        if "No such file" in exc.strerror:
            raise CommandExecutionError(f"File not found: {path}")
        raise CommandExecutionError(f"Unknown Error: {exc.strerror}")

    if not ret:
        return {}

    attrs_ids = ret.split("\n")
    attrs = {}

    for id_ in attrs_ids:
        attrs[id_] = read(path, id_, **{"hex": hex_})

    return attrs


def read(path, attribute, **kwargs):
    """
    Read the given attributes on the given file/directory

    :param str path: The file to get attributes from

    :param str attribute: The attribute to read

    :param bool hex: Return the values with forced hexadecimal values

    :return: A string containing the value of the named attribute
    :rtype: str

    :raises: CommandExecutionError on file not found, attribute not found, and
        any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' xattr.read /path/to/file com.test.attr
        salt '*' xattr.read /path/to/file com.test.attr hex=True
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    hex_ = kwargs.pop("hex", False)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    cmd = ["xattr", "-p"]
    if hex_:
        cmd.append("-x")
    cmd.extend([attribute, path])

    try:
        ret = salt.utils.mac_utils.execute_return_result(cmd)
    except UnicodeDecodeError as exc:
        # Mimic the builtin xattr tool by replacing undecodeable bytes.
        return exc.object.decode(errors="replace")
    except CommandExecutionError as exc:
        if "No such file" in exc.strerror:
            raise CommandExecutionError(f"File not found: {path}")
        if "No such xattr" in exc.strerror:
            raise CommandExecutionError(f"Attribute not found: {attribute}")
        raise CommandExecutionError(f"Unknown Error: {exc.strerror}")

    return ret


def write(path, attribute, value, **kwargs):
    """
    Causes the given attribute name to be assigned the given value

    :param str path: The file(s) to get attributes from

    :param str attribute: The attribute name to be written to the file/directory

    :param str value: The value to assign to the given attribute

    :param bool hex: Set the values with forced hexadecimal values

    :return: True if successful, otherwise False
    :rtype: bool

    :raises: CommandExecutionError on file not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' xattr.write /path/to/file "com.test.attr" "value"

    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    hex_ = kwargs.pop("hex", False)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    cmd = ["xattr", "-w"]
    if hex_:
        cmd.append("-x")
    cmd.extend([attribute, value, path])

    try:
        salt.utils.mac_utils.execute_return_success(cmd)
    except CommandExecutionError as exc:
        if "No such file" in exc.strerror:
            raise CommandExecutionError(f"File not found: {path}")
        raise CommandExecutionError(f"Unknown Error: {exc.strerror}")

    return read(path, attribute, **{"hex": hex_}) == value


def delete(path, attribute):
    """
    Removes the given attribute from the file

    :param str path: The file(s) to get attributes from

    :param str attribute: The attribute name to be deleted from the
        file/directory

    :return: True if successful, otherwise False
    :rtype: bool

    :raises: CommandExecutionError on file not found, attribute not found, and
        any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' xattr.delete /path/to/file "com.test.attr"
    """
    cmd = f'xattr -d "{attribute}" "{path}"'
    try:
        salt.utils.mac_utils.execute_return_success(cmd)
    except CommandExecutionError as exc:
        if "No such file" in exc.strerror:
            raise CommandExecutionError(f"File not found: {path}")
        if "No such xattr" in exc.strerror:
            raise CommandExecutionError(f"Attribute not found: {attribute}")
        raise CommandExecutionError(f"Unknown Error: {exc.strerror}")

    return attribute not in list_(path)


def clear(path):
    """
    Causes the all attributes on the file/directory to be removed

    :param str path: The file(s) to get attributes from

    :return: True if successful, otherwise False

    :raises: CommandExecutionError on file not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' xattr.delete /path/to/file "com.test.attr"
    """
    cmd = f'xattr -c "{path}"'
    try:
        salt.utils.mac_utils.execute_return_success(cmd)
    except CommandExecutionError as exc:
        if "No such file" in exc.strerror:
            raise CommandExecutionError(f"File not found: {path}")
        raise CommandExecutionError(f"Unknown Error: {exc.strerror}")

    return list_(path) == {}



# -----------------------------------------------------------------------------
#  Copyright (C) 2011-2012 Travis Cline
#
#  This file is part of pyzmq
#  It is adapted from upstream project zeromq_gevent under the New BSD License
#
#  Distributed under the terms of the New BSD License.  The full license is in
#  the file LICENSE.BSD, distributed as part of this software.
# -----------------------------------------------------------------------------

"""zmq.green - gevent compatibility with zeromq.

Usage
-----

Instead of importing zmq directly, do so in the following manner:

..

    import zmq.green as zmq


Any calls that would have blocked the current thread will now only block the
current green thread.

This compatibility is accomplished by ensuring the nonblocking flag is set
before any blocking operation and the ØMQ file descriptor is polled internally
to trigger needed events.
"""

from __future__ import annotations

from typing import List

import zmq as _zmq
from zmq import *
from zmq.green.core import _Context, _Socket
from zmq.green.poll import _Poller

Context = _Context  # type: ignore
Socket = _Socket  # type: ignore
Poller = _Poller  # type: ignore

from zmq.green.device import device  # type: ignore

__all__: list[str] = []
# adding `__all__` to __init__.pyi gets mypy all confused
__all__.extend(_zmq.__all__)  # type: ignore



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers

DESCRIPTOR: _descriptor.FileDescriptor

class Authentication(_message.Message):
    __slots__ = ("rules", "providers")
    RULES_FIELD_NUMBER: _ClassVar[int]
    PROVIDERS_FIELD_NUMBER: _ClassVar[int]
    rules: _containers.RepeatedCompositeFieldContainer[AuthenticationRule]
    providers: _containers.RepeatedCompositeFieldContainer[AuthProvider]
    def __init__(
        self,
        rules: _Optional[_Iterable[_Union[AuthenticationRule, _Mapping]]] = ...,
        providers: _Optional[_Iterable[_Union[AuthProvider, _Mapping]]] = ...,
    ) -> None: ...

class AuthenticationRule(_message.Message):
    __slots__ = ("selector", "oauth", "allow_without_credential", "requirements")
    SELECTOR_FIELD_NUMBER: _ClassVar[int]
    OAUTH_FIELD_NUMBER: _ClassVar[int]
    ALLOW_WITHOUT_CREDENTIAL_FIELD_NUMBER: _ClassVar[int]
    REQUIREMENTS_FIELD_NUMBER: _ClassVar[int]
    selector: str
    oauth: OAuthRequirements
    allow_without_credential: bool
    requirements: _containers.RepeatedCompositeFieldContainer[AuthRequirement]
    def __init__(
        self,
        selector: _Optional[str] = ...,
        oauth: _Optional[_Union[OAuthRequirements, _Mapping]] = ...,
        allow_without_credential: bool = ...,
        requirements: _Optional[_Iterable[_Union[AuthRequirement, _Mapping]]] = ...,
    ) -> None: ...

class JwtLocation(_message.Message):
    __slots__ = ("header", "query", "cookie", "value_prefix")
    HEADER_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    COOKIE_FIELD_NUMBER: _ClassVar[int]
    VALUE_PREFIX_FIELD_NUMBER: _ClassVar[int]
    header: str
    query: str
    cookie: str
    value_prefix: str
    def __init__(
        self,
        header: _Optional[str] = ...,
        query: _Optional[str] = ...,
        cookie: _Optional[str] = ...,
        value_prefix: _Optional[str] = ...,
    ) -> None: ...

class AuthProvider(_message.Message):
    __slots__ = (
        "id",
        "issuer",
        "jwks_uri",
        "audiences",
        "authorization_url",
        "jwt_locations",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    ISSUER_FIELD_NUMBER: _ClassVar[int]
    JWKS_URI_FIELD_NUMBER: _ClassVar[int]
    AUDIENCES_FIELD_NUMBER: _ClassVar[int]
    AUTHORIZATION_URL_FIELD_NUMBER: _ClassVar[int]
    JWT_LOCATIONS_FIELD_NUMBER: _ClassVar[int]
    id: str
    issuer: str
    jwks_uri: str
    audiences: str
    authorization_url: str
    jwt_locations: _containers.RepeatedCompositeFieldContainer[JwtLocation]
    def __init__(
        self,
        id: _Optional[str] = ...,
        issuer: _Optional[str] = ...,
        jwks_uri: _Optional[str] = ...,
        audiences: _Optional[str] = ...,
        authorization_url: _Optional[str] = ...,
        jwt_locations: _Optional[_Iterable[_Union[JwtLocation, _Mapping]]] = ...,
    ) -> None: ...

class OAuthRequirements(_message.Message):
    __slots__ = ("canonical_scopes",)
    CANONICAL_SCOPES_FIELD_NUMBER: _ClassVar[int]
    canonical_scopes: str
    def __init__(self, canonical_scopes: _Optional[str] = ...) -> None: ...

class AuthRequirement(_message.Message):
    __slots__ = ("provider_id", "audiences")
    PROVIDER_ID_FIELD_NUMBER: _ClassVar[int]
    AUDIENCES_FIELD_NUMBER: _ClassVar[int]
    provider_id: str
    audiences: str
    def __init__(
        self, provider_id: _Optional[str] = ..., audiences: _Optional[str] = ...
    ) -> None: ...
